# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

"""Triton implementation of Flash Attention w/ Paged KV Cache + FlashDecoding.

Compatible with A10, H100, AMD MI300X.
"""

import torch
import triton
import triton.language as tl
from triton.language.extra import libdevice  # type: ignore[attr-defined]

from conch.platforms import current_platform


@triton.jit  # type: ignore[misc]
def _paged_attention_compute_splits_kernel(  # noqa: PLR0913, PLR0915
    # Pointers to tensors
    output_scratchpad_ptr: tl.tensor,  # (batch_size, max_num_blocks_per_sequence, num_query_heads, head_size)
    lse_scratchpad_ptr: tl.tensor,  # (batch_size, max_num_blocks_per_sequence, num_query_heads)
    query_ptr: tl.tensor,  # (batch_size, num_query_heads, head_size)
    key_cache_ptr: tl.tensor,  # (num_cache_blocks, cache_block_size, num_kv_heads, head_size)
    value_cache_ptr: tl.tensor,  # (num_cache_blocks, cache_block_size, num_kv_heads, head_size)
    block_table_ptr: tl.tensor,  # (batch_size, max_num_blocks_per_sequence)
    seq_lens_ptr: tl.tensor,  # (batch_size, )
    k_scale_ptr: tl.tensor,  # (1,)
    v_scale_ptr: tl.tensor,  # (1,)
    # Scalar arguments
    scale: float,
    num_cache_blocks_per_split: int,
    softcap: float,
    # Sizes of tensors above
    head_size: int,  # output.shape[3]
    query_group_size: int,  # num_query_heads // num_kv_heads
    # Strides for tensors above
    output_scratchpad_batch_stride: int,  # output_scratchpad.stride(0)
    output_scratchpad_cache_block_stride: int,  # output_scratchpad.stride(1)
    output_scratchpad_head_stride: int,  # output_scratchpad.stride(2)
    lse_scratchpad_batch_stride: int,  # lse_scratchpad.stride(0)
    lse_scratchpad_cache_block_stride: int,  # lse_scratchpad.stride(1)
    query_batch_stride: int,  # query.stride(0)
    query_head_stride: int,  # query.stride(1)
    kv_page_stride: int,  # key_cache.stride(0), same for key and value
    kv_cache_block_stride: int,  # key_cache.stride(1), same for key and value
    kv_head_stride: int,  # key_cache.stride(2), same for key and value
    kv_head_element_stride: int,  # key_cache.stride(3), same for key and value
    block_table_batch_stride: int,  # block_table.stride(0)
    # Constexprs
    cxpr_cache_block_size: tl.constexpr,
    cxpr_head_size_padded: tl.constexpr,
    cxpr_query_group_size_padded: tl.constexpr,
    cxpr_is_softcap: tl.constexpr,
    cxpr_apply_fp8_scaling: tl.constexpr,
    cxpr_is_rocm: tl.constexpr,
) -> None:
    """PagedAttention kernel: compute attention for a split block.

    Args:
        output_scratchpad_ptr: Pointer to tensor as scratchpad for output of each cache block, shape: (batch_size, max_num_blocks_per_sequence, num_query_heads, head_size).
        lse_scratchpad_ptr: Pointer to tensor as scratchpad for log-sum-exp of each cache block, shape: (batch_size, max_num_blocks_per_sequence, num_query_heads).
        query_ptr: Pointer to tensor storing the query, shape: (batch_size, num_query_heads, head_size).
        key_cache_ptr: Tensor with cached K values, shape: (num_blocks, cache_block_size, num_kv_heads, head_size).
        value_cache_ptr: Tensor with cached V values, shape: (num_blocks, cache_block_size, num_kv_heads, head_size).
        block_table_ptr: Pointer to tensor storing the mapping from batch to cache blocks, shape: (batch_size, max_num_blocks_per_sequence).
        seq_lens_ptr: Pointer to tensor holding the current sequence length for each sequence in the batch, shape: (batch_size, ).
        k_scale_ptr: Pointer to tensor holding fp8 scaling factor for k.
        v_scale_ptr: Pointer to tensor holding fp8 scaling factor for v.
        scale: Scaling factor, 1/sqrt(head_size).
        num_cache_blocks_per_split: The maximum number of cache blocks in each split (max num each kernel will process).
        softcap: Logits softcap to apply.
        head_size: Actual head dim, not padded to power-of-two.
        query_group_size: The number of query heads to group together, not padded to power-of-two.
        output_scratchpad_batch_stride: Stride of the output scratchpad tensor in the 0th dimension.
        output_scratchpad_cache_block_stride: Stride of the output scratchpad tensor in the 1st dimension.
        output_scratchpad_head_stride: Stride of the output scratchpad tensor in the 2nd dimension.
        lse_scratchpad_batch_stride: Stride of the log-sum-exp scratchpad tensor in the 0th dimension.
        lse_scratchpad_cache_block_stride: Stride of the log-sum-exp scratchpad tensor in the 1st dimension.
        query_batch_stride: Stride of the query tensor in the 0th dimension.
        query_head_stride: Stride of the query tensor in the 1st dimension.
        kv_page_stride: Stride of the k/v tensors in the 0th dimension.
        kv_cache_block_stride: Stride of the k/v tensors in the 1st dimension.
        kv_head_stride: Stride of the k/v tensors in the 2nd dimension.
        kv_head_element_stride: Stride of the k/v tensors in the 3rd dimension.
        block_table_batch_stride: Stride of the block table tensor in the 0th dimension.
        cxpr_cache_block_size: The size of the cache blocks (must be power of two!).
        cxpr_head_size_padded: The head size of the attention layer padded to the next power of two.
        cxpr_query_group_size_padded: The query group size padded to the next power of two.
        cxpr_is_softcap: Whether or not logits softcapping will be applied.
        cxpr_apply_fp8_scaling: Whether or not to apply FP8 scaling.
        cxpr_is_rocm: Whether or not we're on AMD.
    """
    # What batch is this program processing?
    batch_index = tl.program_id(0)
    # What "split" of the overall data (between 1 and N cache blocks) is this program processing?
    split_index = tl.program_id(1)
    # What KV head is this program processing?
    kv_head_index = tl.program_id(2)

    # Get type that we should be using for accumulating results/intermediate calculations
    dtype = output_scratchpad_ptr.dtype.element_ty

    # Load scalar current_sequence_length for the current sequence in the batch
    current_sequence_length = tl.load(seq_lens_ptr + batch_index)

    # The length of the current sequence will tell us how many cache blocks we need to read
    current_seq_num_cache_blocks = tl.cdiv(current_sequence_length, cxpr_cache_block_size)

    # What is the first cache block index in the set for this split
    starting_cache_block_index = split_index * num_cache_blocks_per_split

    # We launch the same number of splits for all sequences in the batch, so different kernel launches will have different numbers of cache blocks
    # to process. So we may have a case where one sequence in a batch has sufficiently fewer cache blocks to process such that a kernel doesn't have
    # any blocks to process.
    if starting_cache_block_index >= current_seq_num_cache_blocks:
        return

    # Offsets to each element of the padded-to-next-power-of-two head size
    head_offsets = tl.arange(0, cxpr_head_size_padded)
    # Mask to only read valid indices of the actual head size
    head_mask = head_offsets < head_size

    # Offsets for each query vector in the group
    query_group_offsets = tl.arange(0, cxpr_query_group_size_padded)
    # Mask out query heads that are just for padding
    query_group_mask = query_group_offsets < query_group_size

    # Offsets for the queries in this block
    query_offsets = query_group_offsets[:, None] * query_head_stride + head_offsets[None, :]
    # Mask out query heads that are just for padding
    query_mask = query_group_mask[:, None] & head_mask[None, :]

    # Offsets for query vector this batch/head
    query_batch_index_offset = batch_index * query_batch_stride
    query_head_index_offset = (kv_head_index * query_group_size) * query_head_stride

    # Load queries for all of the query heads that correspond to this KV head
    query = tl.load(
        query_ptr + query_batch_index_offset + query_head_index_offset + query_offsets, mask=query_mask, other=0.0
    )

    # Offset for the current kv_head in the key_cache and value_cache
    kv_head_index_offset = kv_head_index * kv_head_stride

    # Pointer arithmetic to get to the entry in the block_table for the current batch_index
    current_block_table_offset = batch_index * block_table_batch_stride
    current_block_table_ptr = block_table_ptr + current_block_table_offset

    # Scratchpad for output from this group of cache blocks
    output = tl.zeros([cxpr_query_group_size_padded, cxpr_head_size_padded], dtype=dtype)
    # Keep running max of softmax numerator (scale * Q * K)
    m_i = tl.full([cxpr_query_group_size_padded], -float("inf"), dtype=dtype)
    # Keep running denominator of softmax
    l_i = tl.full([cxpr_query_group_size_padded], 0.0, dtype=dtype)

    # Offsets for each element of the cache block
    cache_block_offsets = tl.arange(0, cxpr_cache_block_size)

    # Iterate through the cache blocks that this kernel is assigned to
    for relative_cache_block_index in range(num_cache_blocks_per_split):
        # Get the actual index of the cache block
        cache_block_index = starting_cache_block_index + relative_cache_block_index

        # If our cache block index is greater than or equal to the number of active cache blocks for the current sequence,
        # skip this block
        # Note: `break` / `continue` is unsupported in Triton, so we have to nest here
        if cache_block_index < current_seq_num_cache_blocks:
            # Calculate number of entries in this cache block (will be a value between 1 and cache_block_size)
            num_entries_in_cache_block = min(
                current_sequence_length - (cache_block_index * cxpr_cache_block_size), cxpr_cache_block_size
            )

            cache_block_mask = cache_block_offsets < num_entries_in_cache_block

            # Offset from the block_table row for the current batch by the number of cache blocks
            current_cache_block_number_ptr = current_block_table_ptr + cache_block_index
            physical_cache_block_number = tl.load(current_cache_block_number_ptr)

            # Calculate address of current cache block
            kv_cache_block_index_offset = physical_cache_block_number * kv_page_stride

            # Load the key block as (cxpr_head_size_padded, cache_block_size)
            # Note: we're loading it transposed here
            key_block_offsets = (
                cache_block_offsets[None, :] * kv_cache_block_stride + kv_head_index_offset + head_offsets[:, None]
            )

            key_block_mask = head_mask[:, None] & cache_block_mask[None, :]

            key_block = tl.load(
                key_cache_ptr + kv_cache_block_index_offset + key_block_offsets, mask=key_block_mask, other=0.0
            )

            if cxpr_apply_fp8_scaling:
                # Dequantize (multiply by scale factor)
                fp8_dtype = tl.float8e4b8 if cxpr_is_rocm else tl.float8e4nv
                k_scale = tl.load(k_scale_ptr)
                key_block = (key_block.to(fp8_dtype, bitcast=True) * k_scale).to(dtype)

            # Multiply query vector by key matrix for this cache block (and apply scaling factor)
            # query.shape -> (query_group_size, head_size)
            # key_block.shape -> (head_size, cache_block_size)
            # qk.shape -> (query_group_size, cache_block_size)
            qk = (scale * tl.dot(query, key_block)).to(dtype)

            # Need to mask out any elements that represent unused cache block entries or padding elements
            # cache_block_mask = tl.arange(0, cxpr_cache_block_size) < num_entries_in_cache_block
            qk_mask = query_group_mask[:, None] & cache_block_mask[None, :]

            # Set masked out elements to -inf
            qk = tl.where(qk_mask, qk, -float("inf")).to(dtype)

            # Handle softcapping
            if cxpr_is_softcap:
                # tanh can only accept fp32 or fp64 arguments
                qk = (softcap * libdevice.tanh((qk / softcap).to(tl.float32))).to(dtype)

            # Reduce maximum between running max and the max of (scale * Q * K) for this cache block
            m_ij = tl.maximum(m_i, tl.max(qk, axis=1)).to(dtype)

            # Calculate numerator of softmax for this cache block
            p = tl.exp((qk - m_ij[:, None]).to(tl.float32)).to(dtype)
            # Need to mask out any elements that represent unused cache block entries or padding elements
            p = tl.where(qk_mask, p, 0.0).to(dtype)

            # Calculate sum of softmax numerator for this cache block
            l_ij = tl.sum(p, axis=1).to(dtype)

            # Calculate correction factor for this cache block
            alpha = tl.exp((m_i - m_ij).to(tl.float32)).to(dtype)

            # Apply scaling factor to running output
            output *= alpha[:, None]

            # Load the value block as (cache_block_size, cxpr_head_size_padded)
            value_block_offsets = (
                cache_block_offsets[:, None] * kv_cache_block_stride + kv_head_index_offset + head_offsets[None, :]
            )

            value_block_mask = cache_block_mask[:, None] & head_mask[None, :]

            value_block = tl.load(
                value_cache_ptr + kv_cache_block_index_offset + value_block_offsets, mask=value_block_mask, other=0.0
            )

            if cxpr_apply_fp8_scaling:
                # Dequantize (multiply by scale factor)
                fp8_dtype = tl.float8e4b8 if cxpr_is_rocm else tl.float8e4nv
                v_scale = tl.load(v_scale_ptr)
                value_block = (value_block.to(fp8_dtype, bitcast=True) * v_scale).to(dtype)

            # Multiply softmax probabilities by value matrix for this cache block
            # p.shape -> (query_group_size, cache_block_size)
            # value_block.shape -> (cache_block_size, head_size)
            # output.shape -> (query_group_size, head_size)
            output += tl.dot(p, value_block).to(dtype)

            # Update running max
            m_i = m_ij
            # Update running denominator
            l_i = l_i * alpha + l_ij

    # Apply correction for denominator of these cache blocks
    output /= l_i[:, None]

    # Calculate offsets to store the output for this head/batch/split
    output_batch_index_offset = batch_index * output_scratchpad_batch_stride
    output_split_index_offset = split_index * output_scratchpad_cache_block_stride
    output_head_index_offset = (kv_head_index * query_group_size) * output_scratchpad_head_stride

    # 2D block of shape (query_group_size_padded, head_size_padded)
    scratch_offsets = query_group_offsets[:, None] * output_scratchpad_head_stride + head_offsets[None, :]
    output_offsets = output_batch_index_offset + output_split_index_offset + output_head_index_offset + scratch_offsets

    # Store output scratchpad results
    tl.store(output_scratchpad_ptr + output_offsets, output, mask=query_mask)

    # Calculate scratchpad log(sum(exp))
    # Note: log() only accepts fp32/fp64 arguments
    lse = m_i + tl.log(l_i.to(tl.float32)).to(dtype)

    # Calculate offsets to store log-sum-exp for this head/batch/split
    lse_scratch_batch_index_offset = batch_index * lse_scratchpad_batch_stride
    lse_scratch_split_index_offset = split_index * lse_scratchpad_cache_block_stride
    lse_scratch_head_index_offsets = kv_head_index * query_group_size
    lse_scratch_offsets = (
        lse_scratch_batch_index_offset
        + lse_scratch_split_index_offset
        + lse_scratch_head_index_offsets
        + query_group_offsets
    )

    # Store lse scratchpad results
    tl.store(lse_scratchpad_ptr + lse_scratch_offsets, lse, mask=query_group_mask)


@triton.jit  # type: ignore[misc]
def _paged_attention_reduce_splits_kernel(  # noqa: PLR0913
    # Pointers to tensors
    output_ptr: tl.tensor,  # (batch_size, num_query_heads, head_size)
    output_scratchpad_ptr: tl.tensor,  # (batch_size, max_num_blocks_per_sequence, num_query_heads, head_size)
    lse_scratchpad_ptr: tl.tensor,  # (batch_size, max_num_blocks_per_sequence, num_query_heads)
    seq_lens_ptr: tl.tensor,  # (batch_size, )
    # Scalars
    num_cache_blocks_per_split: int,
    # Sizes of tensors above
    head_size: int,  # output.shape[2]
    # Strides for tensors above
    output_batch_stride: int,  # output.stride(0)
    output_head_stride: int,  # output.stride(1)
    output_scratchpad_batch_stride: int,  # output_scratchpad.stride(0)
    output_scratchpad_split_stride: int,  # output_scratchpad.stride(1)
    output_scratchpad_head_stride: int,  # output_scratchpad.stride(2)
    lse_scratchpad_batch_stride: int,  # lse_scratchpad.stride(0)
    lse_scratchpad_split_stride: int,  # lse_scratchpad.stride(1)
    # Constexprs
    cxpr_cache_block_size: tl.constexpr,
    cxpr_head_size_padded: tl.constexpr,
) -> None:
    """PagedAttention kernel: reduce results across all splits.

    Args:
        output_ptr: Pointer to tensor for final output, shape: (batch_size, num_query_heads, head_size).
        output_scratchpad_ptr: Pointer to tensor as scratchpad for output of each cache block, shape: (batch_size, max_num_blocks_per_sequence, num_query_heads, head_size).
        lse_scratchpad_ptr: Pointer to tensor as scratchpad for log-sum-exp of each cache block, shape: (batch_size, max_num_blocks_per_sequence, num_query_heads).
        seq_lens_ptr: Pointer to tensor holding the current sequence length for each sequence in the batch, shape: (batch_size, ).
        num_cache_blocks_per_split: The maximum number of cache blocks each split will process.
        head_size: Actual head dim, not padded to power-of-two.
        output_batch_stride: Stride of the output tensor in the 0th dimension.
        output_head_stride: Stride of the output tensor in the 1st dimension.
        output_scratchpad_batch_stride: Stride of the output scratchpad tensor in the 0th dimension.
        output_scratchpad_split_stride: Stride of the output scratchpad tensor in the 1st dimension.
        output_scratchpad_head_stride: Stride of the output scratchpad tensor in the 2nd dimension.
        lse_scratchpad_batch_stride: Stride of the log-sum-exp scratchpad tensor in the 0th dimension.
        lse_scratchpad_split_stride: Stride of the log-sum-exp scratchpad tensor in the 1st dimension.
        cxpr_cache_block_size: The size of the cache blocks (must be power of two!), as constexpr so that we can use for reshaping tensors.
        cxpr_head_size_padded: The head size of the attention layer padded to the next power of two.
    """
    # What batch is this program processing?
    batch_index = tl.program_id(0)
    # What head is this program processing?
    head_index = tl.program_id(1)

    # Get type that we should be using for accumulating results/intermediate calculations
    dtype = output_ptr.dtype.element_ty

    # Accumulator for the output of this batch/head
    output = tl.zeros((cxpr_head_size_padded,), dtype=dtype)
    # Running max of block lse
    # Note: this syntax lets you create a scalar of a specific datatype, see:
    # https://github.com/triton-lang/triton/issues/2939#issuecomment-1892525932
    m_i = tl.full([], -float("inf"), dtype=dtype)
    # Running final scale factor
    l_i = tl.full([], 0.0, dtype=dtype)

    # Load scalar current_sequence_length for the current batch
    current_sequence_length = tl.load(seq_lens_ptr + batch_index)

    # The length of the current sequence will tell us how many cache blocks we need to read
    current_seq_num_cache_blocks = tl.cdiv(current_sequence_length, cxpr_cache_block_size)

    # Offsets to each element of the padded-to-next-power-of-two head size
    head_offsets = tl.arange(0, cxpr_head_size_padded)
    # Mask to only read valid indices of the actual head size
    head_mask = head_offsets < head_size

    num_splits_this_seq = tl.cdiv(current_seq_num_cache_blocks, num_cache_blocks_per_split)

    # Iterate through every cache block for the current sequence
    for split_index in range(num_splits_this_seq):
        # Calculate offsets to load the output scratchpad for this head/batch/cache block
        output_scratchpad_batch_index_offset = batch_index * output_scratchpad_batch_stride
        output_scratchpad_split_index_offset = split_index * output_scratchpad_split_stride
        output_scratchpad_head_index_offset = head_index * output_scratchpad_head_stride
        output_scratchpad_offsets = (
            output_scratchpad_batch_index_offset
            + output_scratchpad_split_index_offset
            + output_scratchpad_head_index_offset
            + head_offsets
        )

        # Load output for this cache block, shape -> (cxpr_head_size_padded,)
        block_output = tl.load(output_scratchpad_ptr + output_scratchpad_offsets, mask=head_mask, other=0.0)

        # Calculate offsets to store log-sum-exp for this head/batch/cache block
        lse_scratch_batch_index_offset = batch_index * lse_scratchpad_batch_stride
        lse_scratch_split_index_offset = split_index * lse_scratchpad_split_stride
        # Load log-sum-exp for this cache block, shape -> scalar
        block_lse = tl.load(
            lse_scratchpad_ptr + lse_scratch_batch_index_offset + lse_scratch_split_index_offset + head_index
        )

        # Reduce running max lse
        m_ij = tl.maximum(m_i, block_lse).to(dtype)

        # Calculate correction factor from previous cache blocks
        # Note: exp() only accepts fp32/fp64 arguments
        alpha = tl.exp((m_i - m_ij).to(tl.float32)).to(dtype)
        # Apply correction factor
        output *= alpha

        # Calculate correction factor from this cache block
        # Note: exp() only accepts fp32/fp64 arguments
        beta = tl.exp((block_lse - m_ij).to(tl.float32)).to(dtype)
        # Apply second correction factor and accumulate running output
        output += (beta * block_output).to(dtype)

        # Update running max
        m_i = m_ij
        # Update running final scale factor
        l_i = l_i * alpha + beta

    # Apply final correction to output
    output /= l_i

    # Calculate offsets to store the output for this head/batch
    output_batch_index_offset = batch_index * output_batch_stride
    output_head_index_offset = head_index * output_head_stride
    output_offsets = output_batch_index_offset + output_head_index_offset + head_offsets
    # Store final result for this head/batch
    tl.store(output_ptr + output_offsets, output, mask=head_mask)


def paged_attention_launcher(  # noqa: PLR0913
    out: torch.Tensor,
    query: torch.Tensor,
    key_cache: torch.Tensor,
    value_cache: torch.Tensor,
    output_scratchpad: torch.Tensor,
    lse_scratchpad: torch.Tensor,
    block_table: torch.Tensor,
    seq_lens: torch.Tensor,
    scale: float | None = None,
    softcap: float = 0.0,
    kv_cache_dtype: str = "auto",
    k_scale: torch.Tensor | None = None,
    v_scale: torch.Tensor | None = None,
) -> None:
    """PagedAttention kernel launcher.

    Args:
        out: Tensor to write the output of the attention calculation, shape: (batch_size, num_heads, head_size).
        query: Query tensor, shape: (batch_size, num_heads, head_size).
        key_cache: Tensor with cached K values, shape: (num_blocks, cache_block_size, num_kv_heads, head_size).
        value_cache: Tensor with cached V values, shape: (num_blocks, cache_block_size, num_kv_heads, head_size).
        output_scratchpad: Tensor used as scratchpad to share cache block outputs between two stages, shape: (batch_size, max_num_blocks_per_sequence, num_query_heads, head_size)
        lse_scratchpad: Tensor used as scratchpad to share cache block log-sum-exp between two stages, shape: (batch_size, max_num_blocks_per_sequence, num_query_heads)
        block_table: Tensor storing the mapping from batch to cache blocks, shape: (batch_size, max_num_blocks_per_sequence).
        seq_lens: Tensor with the sequence length of each index in the batch, shape: (batch_size, ).
        scale: Scaling factor, 1/sqrt(head_size).
        softcap: Logit softcap to apply (0.0 means no softcap will be applied).
        kv_cache_dtype: If this dtype is fp8, apply scaling.
        k_scale: Fp8 scaling factor for k.
        v_scale: Fp8 scaling factor for v.
    """
    assert query.shape == out.shape  # noqa: S101
    assert key_cache.shape == value_cache.shape  # noqa: S101
    assert key_cache.stride(0) == value_cache.stride(0)  # noqa: S101
    assert key_cache.stride(1) == value_cache.stride(1)  # noqa: S101
    assert key_cache.stride(2) == value_cache.stride(2)  # noqa: S101
    assert key_cache.stride(3) == value_cache.stride(3)  # noqa: S101
    assert key_cache.stride(3) == 1  # noqa: S101
    assert softcap >= 0.0  # noqa: S101

    allowed_in_out_dtypes = [torch.float32, torch.float16, torch.bfloat16]
    assert query.dtype in allowed_in_out_dtypes  # noqa: S101
    assert out.dtype == query.dtype  # noqa: S101
    assert output_scratchpad.dtype == query.dtype  # noqa: S101
    assert lse_scratchpad.dtype == query.dtype  # noqa: S101

    # Perform unchecked size accesses, assume has already been checked
    batch_size, num_query_heads, head_size = out.shape
    num_cache_blocks, cache_block_size, num_kv_heads, _ = key_cache.shape
    _, max_num_blocks_per_sequence = block_table.shape
    _, max_num_splits, _, _ = output_scratchpad.shape

    assert cache_block_size == triton.next_power_of_2(cache_block_size), "Cache block size must be a power of two!"  # noqa: S101

    # Need sizes to be constexpr in order to reshape tensors in kernel
    cxpr_cache_block_size: tl.constexpr = cache_block_size
    cxpr_head_size_padded: tl.constexpr = triton.next_power_of_2(head_size)

    # For parity with Dao Flash Attention, softcap == 0.0 means no softcapping
    cxpr_is_softcap: tl.constexpr = softcap > 0.0

    cxpr_apply_fp8_scaling: tl.constexpr = kv_cache_dtype == "fp8" or kv_cache_dtype == "fp8_e4m3"
    cxpr_is_rocm: tl.constexpr = current_platform.is_amd()

    # How many query heads correspond to the same KV head?
    query_group_size = num_query_heads // num_kv_heads
    # We pad this size to be at least 16 so that we can use `tl.dot()` operations inside of the kernel
    cxpr_query_group_size_padded = max(16, triton.next_power_of_2(query_group_size))

    # What is the maximum number of stage 1 kernels to launch per batch/head?
    # Each kernel processes up to {cache_block_size} tokens at a time (in many cases cache_block_size=32 for vLLM), so we can process
    # a sequence up to {max_num_splits * cache_block_size} == 64 * 32 == 2048 tokens before a stage 1 kernel will process multiple cache
    # blocks. This helps to reduce the overhead of kernel launches / split reduction for long sequences.
    # Note: we may need to tune this value for a given HW platform.
    num_splits = min(max_num_blocks_per_sequence, max_num_splits)

    num_cache_blocks_per_split = triton.cdiv(max_num_blocks_per_sequence, num_splits)

    # Use default scaling factor if not provided
    if scale is None:
        scale = float(1.0 / (head_size**0.5))

    if cxpr_apply_fp8_scaling:
        assert k_scale is not None  # noqa: S101
        assert v_scale is not None  # noqa: S101
        assert k_scale.numel() == 1  # noqa: S101
        assert v_scale.numel() == 1  # noqa: S101

    # For computing attention for split block (stage 1): parallelize over batches, cache blocks, and KV heads.
    # Note: if the number of cache blocks in a sequence is very large, it is more efficient to handle multiple blocks
    # in one launch of the stage 1 kernel to reduce the overhead of reduction
    stage1_grid = (batch_size, num_splits, num_kv_heads)

    # Launch stage 1 kernel
    _paged_attention_compute_splits_kernel[stage1_grid](
        # Relevant tensors
        output_scratchpad,
        lse_scratchpad,
        query,
        key_cache,
        value_cache,
        block_table,
        seq_lens,
        k_scale,
        v_scale,
        # Scalars
        scale,
        num_cache_blocks_per_split,
        softcap,
        # Sizes of relevant tensors
        head_size,
        query_group_size,
        # Strides of relevant tensors
        output_scratchpad.stride(0),
        output_scratchpad.stride(1),
        output_scratchpad.stride(2),
        lse_scratchpad.stride(0),
        lse_scratchpad.stride(1),
        query.stride(0),
        query.stride(1),
        key_cache.stride(0),
        key_cache.stride(1),
        key_cache.stride(2),
        key_cache.stride(3),
        block_table.stride(0),
        # Constexpr sizes
        cxpr_cache_block_size,
        cxpr_head_size_padded,
        cxpr_query_group_size_padded,
        cxpr_is_softcap,
        cxpr_apply_fp8_scaling,
        cxpr_is_rocm,
    )

    # For reducing over splits (stage 2): parallelize over batches and query heads
    stage2_grid = (batch_size, num_query_heads)

    # Launch stage 2 kernel
    _paged_attention_reduce_splits_kernel[stage2_grid](
        # Relevant tensors
        out,
        output_scratchpad,
        lse_scratchpad,
        seq_lens,
        # Scalars
        num_cache_blocks_per_split,
        # Sizes of relevant tensors
        head_size,
        # Strides of relevant tensors
        out.stride(0),
        out.stride(1),
        output_scratchpad.stride(0),
        output_scratchpad.stride(1),
        output_scratchpad.stride(2),
        lse_scratchpad.stride(0),
        lse_scratchpad.stride(1),
        # Constexpr sizes
        cxpr_cache_block_size,
        cxpr_head_size_padded,
    )



# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

"""Test cases for Triton rms_norm."""

from typing import Final

import pytest
import torch

from conch.ops.normalization.rms_norm import fused_add_rms_norm as fused_add_rms_norm_triton
from conch.ops.normalization.rms_norm import rms_norm as rms_norm_triton
from conch.platforms import current_platform
from conch.reference.normalization.rms_norm import fused_add_rms_norm as fused_add_rms_norm_reference
from conch.reference.normalization.rms_norm import rms_norm as rms_norm_reference
from conch.third_party.vllm.utils import seed_everything

_DTYPES: Final = [torch.float16, torch.bfloat16, torch.float32]
_HIDDEN_SIZES: Final = [855, 1024]
_NUM_TOKENS: Final = [80, 256]
_EPSILONS: Final = [1e-6, 1e-4]


@pytest.mark.parametrize("dtype", _DTYPES)
@pytest.mark.parametrize("hidden_size", _HIDDEN_SIZES)
@pytest.mark.parametrize("num_tokens", _NUM_TOKENS)
@pytest.mark.parametrize("epsilon", _EPSILONS)
def test_rms_norm(dtype: torch.dtype, hidden_size: int, num_tokens: int, epsilon: float) -> None:
    """Test Triton rms_norm implementation."""
    seed: Final = 0
    seed_everything(seed)

    device: Final = torch.device(current_platform.device)
    torch.set_default_device(device)

    x_shape = (num_tokens, hidden_size)
    x = torch.randn(x_shape, dtype=dtype, device=device)
    weight = torch.randn((hidden_size,), dtype=dtype, device=device)

    triton_output = rms_norm_triton(x, weight, epsilon)
    reference_output = rms_norm_reference(x, weight, epsilon)

    tolerance = 1e-2
    torch.testing.assert_close(triton_output, reference_output, rtol=tolerance, atol=tolerance)


@pytest.mark.parametrize("dtype", _DTYPES)
@pytest.mark.parametrize("hidden_size", _HIDDEN_SIZES)
@pytest.mark.parametrize("num_tokens", _NUM_TOKENS)
@pytest.mark.parametrize("epsilon", _EPSILONS)
def test_fused_add_rms_norm(dtype: torch.dtype, hidden_size: int, num_tokens: int, epsilon: float) -> None:
    """Test Triton fused_add_rms_norm implementation."""
    seed: Final = 0
    seed_everything(seed)

    device: Final = torch.device(current_platform.device)
    torch.set_default_device(device)

    x_shape = (num_tokens, hidden_size)
    x = torch.randn(x_shape, dtype=dtype, device=device)
    residual = torch.randn(x_shape, dtype=dtype, device=device)
    weight = torch.randn((hidden_size,), dtype=dtype, device=device)

    # We need to clone the tensors because the fused_add_rms_norm kernel modifies them in-place
    triton_x = x.clone()
    reference_x = x.clone()
    triton_residual = residual.clone()
    reference_residual = residual.clone()

    triton_output, triton_residual = fused_add_rms_norm_triton(triton_x, triton_residual, weight, epsilon)
    reference_output, reference_residual = fused_add_rms_norm_reference(
        reference_x, reference_residual, weight, epsilon
    )

    tolerance = 1e-2
    torch.testing.assert_close(triton_output, reference_output, rtol=tolerance, atol=tolerance)
    torch.testing.assert_close(triton_residual, reference_residual, rtol=tolerance, atol=tolerance)



# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

"""Plot benchmarking results."""

from pathlib import Path

import click
import pandas as pd  # type: ignore[import-untyped]
from matplotlib import pyplot as plt


@click.command()
@click.option(
    "--results-directory",
    required=True,
    type=click.Path(path_type=Path, exists=True, dir_okay=True, file_okay=False),
    help="Path to the benchmark results directory.",
)
@click.option(
    "--x-axis",
    required=True,
    type=str,
    help="Name of the column to use for the x-axis.",
)
@click.option(
    "--y-axis",
    required=True,
    type=str,
    default="runtime_ms",
    help="Name of the column to use for the y-axis.",
)
@click.option(
    "--triton-tag",
    required=True,
    type=str,
    default="Triton",
    help="The 'tag' column value for Triton results.",
)
@click.option(
    "--baseline-tag",
    required=True,
    type=str,
    default="Baseline",
    help="The 'tag' column value for baseline results.",
)
def main(results_directory: Path, x_axis: str, y_axis: str, triton_tag: str, baseline_tag: str) -> None:
    """Main function to plot benchmarking results."""
    dataframes = [pd.read_csv(file) for file in results_directory.iterdir() if file.suffix == ".csv"]
    if not dataframes:
        print(f"No CSV files found in the specified directory ({results_directory = }).")
        return

    if not all(all(df.columns == dataframes[0].columns) for df in dataframes):
        print("CSV files have different columns. Please ensure they are consistent.")
        return

    df = pd.concat(dataframes, ignore_index=True)

    triton_df = df[df["tag"] == triton_tag].sort_values(by=x_axis)
    baseline_df = df[df["tag"] == baseline_tag].sort_values(by=x_axis)

    columns_to_exclude = ["tag", "platform", x_axis, y_axis]
    metadata_columns = [col for col in triton_df.columns if col not in columns_to_exclude]

    triton_metadata = triton_df[metadata_columns].set_index(metadata_columns[0]).reset_index(drop=True)
    baseline_metadata = baseline_df[metadata_columns].set_index(metadata_columns[0]).reset_index(drop=True)

    if not all(triton_metadata == baseline_metadata):
        print("Metadata does not match between Triton and baseline results.")
        return

    metadata = triton_df[metadata_columns].iloc[0].to_dict()

    # Plot the results
    fig, ax = plt.subplots(figsize=(16, 9))
    plt.suptitle(f"Benchmarking Results: {results_directory.name}")
    plt.title(f"Metadata: {metadata}")
    plt.xlabel(x_axis)
    plt.ylabel(y_axis)

    ax.plot(
        triton_df[x_axis],
        triton_df[y_axis],
        label=f"{triton_tag} (platform={triton_df['platform'].iloc[0]}, n={len(triton_df)})",
        color="blue",
    )

    ax.plot(
        baseline_df[x_axis],
        baseline_df[y_axis],
        label=f"{baseline_tag} (platform={baseline_df['platform'].iloc[0]}, n={len(baseline_df)})",
        color="red",
    )

    box = ax.get_position()
    ax.set_position((box.x0, box.y0, box.width * 0.85, box.height))
    ax.legend(loc="center left", bbox_to_anchor=(1, 0.5))

    output_filename = results_directory / "plot.png"
    plt.savefig(output_filename)
    print(f"Plot saved to '{output_filename}'")


if __name__ == "__main__":
    main()



# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

"""Test Triton PagedAttention."""

import math
from typing import Final

import pytest
import torch
import torch.nn.functional as F  # noqa: N812
from einops import einsum, rearrange

from conch import envs
from conch.ops.attention.paged_attention import paged_attention
from conch.platforms import current_platform
from conch.third_party.vllm.utils import create_tensors, seed_everything

_ENABLE_VLLM: Final = envs.CONCH_ENABLE_VLLM and current_platform.has_cuda()
_BATCH_SIZES: Final = [1, 4, 8]
_CACHE_BLOCK_SIZES_VLLM: Final = [16, 32]
_CACHE_BLOCK_SIZES_FLASH: Final = [16, 32, 64]
_HEAD_SIZES_VLLM: Final = [64, 80, 96, 112, 128, 192, 256]
_HEAD_SIZES_FLASH: Final = [64, 96, 128, 256]
_SEQUENCE_LENGTHS: Final = [16, 128, 240, 333, 1002, 2048]
# MHA, MQA, and GQA
# - MHA: num_query_heads == num_kv_heads
# - MQA: num_kv_heads == 1
# - GQA: num_query_heads != num_kv_heads && num_kv_heads != 1
_NUM_HEADS: Final = [(8, 8), (4, 4), (8, 1), (4, 1), (16, 4), (16, 2), (8, 4), (8, 2)]
# Too many parameterizations makes the SDPA test cases too slow
_BATCH_SIZES_ABRIDGED: Final = [4]
_SEQUENCE_LENGTHS_ABRIDGED: Final = [240, 333, 1002]
_NUM_HEADS_ABRIDGED: Final = [(8, 8), (4, 1), (16, 4)]


def _get_tolerance_for_dtype(dtype: torch.dtype) -> float:
    """Get expected tolerance to match to for a given dtype."""
    if dtype == torch.float16:
        return 5e-3

    if dtype == torch.bfloat16:
        return 3e-2

    if dtype == torch.float32:
        return 2e-3

    msg = f"Unsupported dtype: '{dtype}'"
    raise NotImplementedError(msg)


def _create_paged_cache_from_contiguous_kv(
    key: torch.Tensor,
    value: torch.Tensor,
    num_cache_blocks: int,
    cache_block_size: int,
    dtype: torch.dtype,
) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """Create paged cache from contiguous KV.

    Args:
        key: Key tensor.
        value: Value tensor.
        num_cache_blocks: Number of cache blocks in the kv caches.
        cache_block_size: Size of each cache block.
        dtype: Data type of tensors.

    Returns:
        Tuple of paged key cache, paged value cache, and block tables for traversing caches.
    """
    assert key.shape == value.shape
    assert key.device == value.device

    device: Final = key.device

    batch_size, num_kv_heads, sequence_length, head_size = key.shape
    max_num_pages_per_seq: Final = num_cache_blocks // batch_size
    max_seq_len = max_num_pages_per_seq * cache_block_size
    assert sequence_length <= max_seq_len

    k_cache = torch.empty(num_cache_blocks, cache_block_size, num_kv_heads, head_size, dtype=dtype, device=device)
    v_cache = torch.empty(num_cache_blocks, cache_block_size, num_kv_heads, head_size, dtype=dtype, device=device)
    block_table = torch.zeros((batch_size, max_num_pages_per_seq), dtype=torch.int32, device=device)

    current_page = 0
    current_page_offset = 0

    for batch_index in range(batch_size):
        current_blocks = []
        for seq_index in range(sequence_length):
            if current_page not in current_blocks:
                current_blocks.append(current_page)

            for head_index in range(num_kv_heads):
                current_k = key[batch_index][head_index][seq_index].detach().clone()
                current_v = value[batch_index][head_index][seq_index].detach().clone()

                k_cache[current_page][current_page_offset][head_index] = current_k
                v_cache[current_page][current_page_offset][head_index] = current_v

            current_page_offset += 1

            if current_page_offset == cache_block_size:
                current_page += 1
                current_page_offset = 0

        block_table[batch_index] = torch.Tensor(current_blocks)

        # Next batch should get its own page, so increment page index and reset offset
        if current_page_offset != 0:
            current_page += 1
            current_page_offset = 0

    return k_cache, v_cache, block_table


def _convert_paged_to_contiguous(
    k_cache_paged: torch.Tensor,
    v_cache_paged: torch.Tensor,
    block_table: torch.Tensor,
    num_kv_heads: int,
    seq_len: int,
    head_size: int,
    cache_block_size: int,
) -> tuple[torch.Tensor, torch.Tensor]:
    """Convert paged kv cache to contiguous.

    Args:
        k_cache_paged: Paged key cache.
        v_cache_paged: Paged value cache.
        block_table: Block tables for traversing kv caches.
        num_kv_heads: The number of kv heads.
        seq_len: The max sequence length.
        head_size: The dimension of the attention head.
        cache_block_size: The size of each cache block.

    Returns:
        Tuple of key cache in contiguous memory and value cache in contiguous memory.
    """
    batch_size, _ = block_table.shape

    k_contiguous = torch.empty(
        (batch_size, seq_len, num_kv_heads, head_size), dtype=k_cache_paged.dtype, device=k_cache_paged.device
    )
    v_contiguous = torch.empty(
        (batch_size, seq_len, num_kv_heads, head_size), dtype=v_cache_paged.dtype, device=v_cache_paged.device
    )

    for batch_index in range(batch_size):
        current_seq_len_begin = 0
        current_seq_len_end = current_seq_len_begin + cache_block_size

        this_k = k_contiguous[batch_index]
        this_v = v_contiguous[batch_index]

        these_blocks = block_table[batch_index]

        for relative_block_index, cache_block_index in enumerate(these_blocks):
            current_k_block = k_cache_paged[cache_block_index]
            current_v_block = v_cache_paged[cache_block_index]

            # The end of the current cache block or the end of the sequence, whichever is smaller
            actual_end = min(current_seq_len_end, seq_len)

            # Relative end is the end index in the current cache block
            relative_end = actual_end - (relative_block_index * cache_block_size)

            this_k[current_seq_len_begin:actual_end, :, :] = current_k_block.view(
                cache_block_size,
                num_kv_heads,
                head_size,
            )[:relative_end, :, :]

            this_v[current_seq_len_begin:actual_end, :, :] = current_v_block.view(
                cache_block_size,
                num_kv_heads,
                head_size,
            )[:relative_end, :, :]

            current_seq_len_begin += cache_block_size
            current_seq_len_end += cache_block_size

    return k_contiguous, v_contiguous


def _gqa(query: torch.Tensor, key: torch.Tensor, value: torch.Tensor, scale: float) -> torch.Tensor:
    """Grouped Query Attention."""
    bq, hq, nq, dq = query.shape
    bk, hk, nk, dk = key.shape
    bv, hv, nv, dv = value.shape

    if not (bq == bk == bv and dq == dk == dv):
        msg = f"Expected query, key, and value to have the same batch size (dim=0) and embedding dimension (dim=3), but got query: {query.shape}, key: {key.shape}, and value: {value.shape}."
        raise ValueError(msg)

    if (hk != hv) or (nk != nv):
        msg = f"Expected key and value to have the same size in dimensions 1 and 2, but got key: {key.shape} and value: {value.shape}."
        raise ValueError(msg)

    if hq % hk != 0:
        msg = f"Expected query heads to be a multiple of key/value heads, but got query: {query.shape} and key/value: {key.shape}."
        raise ValueError(msg)

    num_head_groups = hq // hk

    query = rearrange(query, "b (h g) n d -> b g h n d", g=num_head_groups)
    similarity = einsum(query, key, "b g h n d, b h s d -> b g h n s")
    attention = F.softmax(similarity * scale, dim=-1)

    # Apply attention matrix to the value Tensor.
    out = einsum(attention, value, "b g h n s, b h s d -> b g h n d")
    # Move head dimension back to axis 2
    return rearrange(out, "b g h n d -> b (h g) n d")


def _run_paged_vs_sdpa(
    batch_size: int,
    head_size: int,
    num_query_heads: int,
    num_kv_heads: int,
    sequence_length: int,
    cache_block_size: int,
    dtype: torch.dtype,
    gqa_workaround: bool,
) -> None:
    """Run test case comparing our Triton PagedAttention kernel vs PyTorch Scaled Dot Product Attention.

    Args:
        batch_size: Number of batches in input.
        head_size: Head dimension.
        num_query_heads: Number of query heads.
        num_kv_heads: Number of key/value heads.
        sequence_length: Sequence length of input.
        cache_block_size: Number of K/V tensors that can fit in a cache block.
        dtype: Data type of tensors.
        gqa_workaround: Enable workaround for Grouped Query Attention. PyTorch SDPA does not currently
            support GQA so use alternative ground-truth function.
    """
    seed: Final = 0
    seed_everything(seed)

    device: Final = torch.device(current_platform.device)
    torch.set_default_device(device)

    # There are some small discrepancies in our output that I believe are caused by fp32<->fp16 conversions inside of the Triton kernel.
    atol: Final = _get_tolerance_for_dtype(dtype)
    rtol: Final = 1e-3

    scale: Final = float(1.0 / (head_size**0.5))

    cache_blocks_per_seq: Final = math.ceil(sequence_length / cache_block_size)
    total_cache_size: Final = batch_size * cache_blocks_per_seq * cache_block_size
    num_cache_blocks: Final = total_cache_size // cache_block_size

    q = torch.randn(
        batch_size, num_query_heads, sequence_length, head_size, device=device, dtype=dtype, requires_grad=False
    )
    k = torch.randn(
        batch_size, num_kv_heads, sequence_length, head_size, device=device, dtype=dtype, requires_grad=False
    )
    v = torch.randn(
        batch_size, num_kv_heads, sequence_length, head_size, device=device, dtype=dtype, requires_grad=False
    )

    sequence_lengths = torch.full((batch_size,), sequence_length, dtype=torch.int32, device=device)

    key_cache_paged, value_cache_paged, block_table = _create_paged_cache_from_contiguous_kv(
        k, v, num_cache_blocks, cache_block_size, dtype
    )
    kc_duplicate, vc_duplicate = _convert_paged_to_contiguous(
        key_cache_paged,
        value_cache_paged,
        block_table,
        num_kv_heads,
        sequence_length,
        head_size,
        cache_block_size,
    )

    kc_duplicate = kc_duplicate.permute(0, 2, 1, 3)
    vc_duplicate = vc_duplicate.permute(0, 2, 1, 3)

    torch.testing.assert_close(k, kc_duplicate, atol=atol, rtol=rtol)
    torch.testing.assert_close(v, vc_duplicate, atol=atol, rtol=rtol)

    q_sdpa = q.detach().clone()
    k_sdpa = k.detach().clone()
    v_sdpa = v.detach().clone()

    if gqa_workaround:
        # Hand-rolled GQA impl because PyTorch does not support it via scaled_dot_product_attention
        out_sdpa = _gqa(q_sdpa, k_sdpa, v_sdpa, scale)
    else:
        # Default PyTorch SDPA
        out_sdpa = F.scaled_dot_product_attention(q_sdpa, k_sdpa, v_sdpa, scale=scale)

    final_out_sdpa = out_sdpa[:, :, -1, :]

    q_paged = q[:, :, -1, :]

    out_paged = paged_attention(
        q_paged,
        key_cache_paged,
        value_cache_paged,
        block_table,
        sequence_lengths,
        scale=scale,
    )

    torch.testing.assert_close(final_out_sdpa, out_paged, atol=atol, rtol=rtol)


def _triton_vs_vllm_cuda(
    batch_size: int,
    cache_block_size: int,
    head_size: int,
    num_query_heads: int,
    num_kv_heads: int,
    sequence_length: int,
    kv_cache_dtype: str,
    dtype: torch.dtype,
) -> None:
    """Run test case comparing our Triton PagedAttention kernel vs vLLM CUDA PagedAttention.

    Args:
        batch_size: Number of batches in input.
        head_size: Head dimension.
        num_query_heads: Number of query heads.
        num_kv_heads: Number of key/value heads.
        sequence_length: Sequence length of input.
        cache_block_size: Number of K/V tensors that can fit in a cache block.
        kv_cache_dtype: Data type of KV cache.
        dtype: Datatype for tensors.
    """
    seed: Final = 0
    seed_everything(seed)

    device: Final = torch.device(current_platform.device)
    torch.set_default_device(device)

    from vllm._custom_ops import paged_attention_v2 as vllm_paged_attention_v2

    query, key_cache_vllm, value_cache_vllm, key_cache_conch, value_cache_conch, block_table, seq_lens = create_tensors(
        head_size,
        sequence_length,
        cache_block_size,
        batch_size,
        num_query_heads,
        num_kv_heads,
        kv_cache_dtype,
        "cuda",
        dtype,
    )

    scale: Final = float(1.0 / (head_size**0.5))
    atol: Final = 1e-3
    rtol: Final = 1e-3

    k_scale = torch.full((1,), 0.5)
    v_scale = torch.full((1,), 0.5)

    # Run vLLM reference implementation
    output_vllm = torch.empty_like(query)

    max_seq_len = torch.max(seq_lens)

    partition_size = 512
    max_num_partitions = (max_seq_len + partition_size - 1) // partition_size
    tmp_output = torch.empty(
        size=(batch_size, num_query_heads, max_num_partitions, head_size),
        dtype=dtype,
        device=query.device,
    )
    exp_sums = torch.empty(
        size=(batch_size, num_query_heads, max_num_partitions),
        dtype=torch.float32,
        device=query.device,
    )
    max_logits = torch.empty_like(exp_sums)

    vllm_paged_attention_v2(
        output_vllm,
        max_logits,
        exp_sums,
        tmp_output,
        query,
        key_cache_vllm,
        value_cache_vllm,
        num_kv_heads,
        scale,
        block_table,
        seq_lens,
        cache_block_size,
        int(max_seq_len.item()),
        None,
        kv_cache_dtype,
        k_scale,
        v_scale,
    )

    # Run Triton implementation
    output_conch = paged_attention(
        query,
        key_cache_conch,
        value_cache_conch,
        block_table,
        seq_lens,
        scale=scale,
        softcap=0.0,
        kv_cache_dtype=kv_cache_dtype,
        k_scale=k_scale,
        v_scale=v_scale,
    )

    torch.testing.assert_close(output_vllm, output_conch, atol=atol, rtol=rtol)


def _triton_vs_flash_attn(
    batch_size: int,
    cache_block_size: int,
    head_size: int,
    num_query_heads: int,
    num_kv_heads: int,
    sequence_length: int,
    dtype: torch.dtype,
    apply_softcap: bool,
) -> None:
    """Run test case comparing our Triton PagedAttention kernel vs CUDA FlashAttnWithKVCache.

    Args:
        batch_size: Number of batches in input.
        head_size: Head dimension.
        num_query_heads: Number of query heads.
        num_kv_heads: Number of key/value heads.
        sequence_length: Sequence length of input.
        cache_block_size: Number of K/V tensors that can fit in a cache block.
        dtype: Datatype for tensors.
        apply_softcap: Whether or not to apply softcapping.
    """
    seed: Final = 0
    seed_everything(seed)

    device: Final = torch.device(current_platform.device)
    torch.set_default_device(device)

    kv_cache_dtype = "auto"

    from vllm.vllm_flash_attn import flash_attn_with_kvcache  # type: ignore[attr-defined, unused-ignore]

    query, _, _, key_cache, value_cache, block_table, seq_lens = create_tensors(
        head_size,
        sequence_length,
        cache_block_size,
        batch_size,
        num_query_heads,
        num_kv_heads,
        kv_cache_dtype,
        "cuda",
        dtype,
    )

    softcap: float = 0.0
    if apply_softcap:
        # Gemma2 default for self-attention layers
        softcap = 30.0
        # Ensure the values of qk are at least within softcap range.
        query = torch.randn_like(query) * softcap

    scale: Final = float(1.0 / (head_size**0.5))

    # There are some small discrepancies in our output that I believe are caused by fp32<->fp16 conversions inside of the Triton kernel.
    atol: Final = 1e-3 if not apply_softcap else 2e-2
    rtol: Final = 1e-3

    # Run FlashAttnWithKVCache implementation
    query_fa = query.unsqueeze(1)

    output_fa = flash_attn_with_kvcache(
        query_fa,
        key_cache,
        value_cache,
        block_table=block_table,
        cache_seqlens=seq_lens,
        softmax_scale=scale,
        causal=True,
        alibi_slopes=None,
        softcap=softcap,
    )

    output_fa = output_fa.squeeze(1)

    # Run Triton implementation
    output_conch = paged_attention(
        query,
        key_cache,
        value_cache,
        block_table,
        seq_lens,
        scale=scale,
        softcap=softcap,
        kv_cache_dtype=kv_cache_dtype,
    )

    assert torch.allclose(output_fa, output_conch, atol=atol, rtol=rtol)


@pytest.mark.parametrize("batch_size", _BATCH_SIZES_ABRIDGED)
@pytest.mark.parametrize("cache_block_size", _CACHE_BLOCK_SIZES_VLLM)
@pytest.mark.parametrize("head_size", _HEAD_SIZES_VLLM)
@pytest.mark.parametrize(("num_query_heads", "num_kv_heads"), _NUM_HEADS_ABRIDGED)
@pytest.mark.parametrize("sequence_length", _SEQUENCE_LENGTHS_ABRIDGED)
@pytest.mark.parametrize("dtype", [torch.float16, torch.bfloat16, torch.float32])
def test_paged_attention(
    batch_size: int,
    cache_block_size: int,
    head_size: int,
    num_query_heads: int,
    num_kv_heads: int,
    sequence_length: int,
    dtype: torch.dtype,
) -> None:
    """Test PagedAttention Triton kernel with various configurations vs. PyTorch."""
    gqa_workaround = num_query_heads // num_kv_heads not in (1, num_query_heads)

    _run_paged_vs_sdpa(
        batch_size,
        head_size,
        num_query_heads,
        num_kv_heads,
        sequence_length,
        cache_block_size,
        dtype,
        gqa_workaround,
    )


@pytest.mark.skipif(not _ENABLE_VLLM, reason="This test case requires vLLM")
@pytest.mark.parametrize("batch_size", _BATCH_SIZES)
@pytest.mark.parametrize("cache_block_size", _CACHE_BLOCK_SIZES_VLLM)
@pytest.mark.parametrize("head_size", _HEAD_SIZES_VLLM)
@pytest.mark.parametrize(("num_query_heads", "num_kv_heads"), _NUM_HEADS)
@pytest.mark.parametrize("sequence_length", _SEQUENCE_LENGTHS)
@pytest.mark.parametrize("kv_cache_dtype", ["auto", "fp8"])
@pytest.mark.parametrize("dtype", [torch.float16, torch.bfloat16, torch.float32])
def test_triton_vs_vllm_cuda(
    batch_size: int,
    cache_block_size: int,
    head_size: int,
    num_query_heads: int,
    num_kv_heads: int,
    sequence_length: int,
    kv_cache_dtype: str,
    dtype: torch.dtype,
) -> None:
    """Test PagedAttention kernel with various configurations vs. vLLM CUDA PagedAttention."""
    if sequence_length < cache_block_size:
        pytest.skip()

    if kv_cache_dtype == "fp8" and not current_platform.supports_fp8():
        pytest.skip()

    _triton_vs_vllm_cuda(
        batch_size,
        cache_block_size,
        head_size,
        num_query_heads,
        num_kv_heads,
        sequence_length,
        kv_cache_dtype,
        dtype,
    )


@pytest.mark.skipif(not _ENABLE_VLLM, reason="This test case requires vLLM")
@pytest.mark.parametrize("batch_size", _BATCH_SIZES)
@pytest.mark.parametrize("cache_block_size", _CACHE_BLOCK_SIZES_FLASH)
@pytest.mark.parametrize("head_size", _HEAD_SIZES_FLASH)
@pytest.mark.parametrize(("num_query_heads", "num_kv_heads"), _NUM_HEADS)
@pytest.mark.parametrize("sequence_length", _SEQUENCE_LENGTHS)
@pytest.mark.parametrize("apply_softcap", [True, False])
@pytest.mark.parametrize("kv_cache_dtype", ["auto", "fp8"])
@pytest.mark.parametrize("dtype", [torch.float16, torch.bfloat16])
def test_triton_vs_flash_attn(
    batch_size: int,
    cache_block_size: int,
    head_size: int,
    num_query_heads: int,
    num_kv_heads: int,
    sequence_length: int,
    kv_cache_dtype: str,
    apply_softcap: bool,
    dtype: torch.dtype,
) -> None:
    """Test PagedAttention kernel with various configurations vs. CUDA FlashAttnWithKVCache."""
    # Only supported on CUDA
    if current_platform.is_amd():
        pytest.skip()

    if sequence_length < cache_block_size:
        pytest.skip()

    _triton_vs_flash_attn(
        batch_size,
        cache_block_size,
        head_size,
        num_query_heads,
        num_kv_heads,
        sequence_length,
        dtype,
        apply_softcap,
    )



# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

"""Test cases for scaled GEMM."""

from typing import Final

import pytest
import torch

from conch.ops.quantization.gemm import scaled_gemm as scaled_gemm_triton
from conch.platforms import current_platform
from conch.reference.quantization.scaled_gemm import scaled_gemm as scaled_gemm_reference
from conch.third_party.vllm.utils import seed_everything


def _is_floating_point_type(dtype: torch.dtype) -> bool:
    """Check whether a type is floating point."""
    return torch.tensor([1, 1], dtype=dtype).is_floating_point()


def _get_8bit_types() -> list[torch.dtype]:
    types = [torch.int8]
    if current_platform.is_amd() and current_platform.supports_fp8():
        types.append(torch.float8_e4m3fnuz)
    elif current_platform.is_nvidia() and current_platform.supports_fp8():
        types.append(torch.float8_e4m3fn)
    return types


@pytest.mark.parametrize(("m_dim", "k_dim", "n_dim"), [(128, 256, 128), (1024, 1024, 1024), (4096, 2048, 4096)])
@pytest.mark.parametrize("input_dtype", _get_8bit_types())
@pytest.mark.parametrize("output_dtype", [torch.float16, torch.bfloat16])
@pytest.mark.parametrize("use_scalar_scale_a", [True, False])
@pytest.mark.parametrize("use_scalar_scale_b", [True, False])
@pytest.mark.parametrize("use_bias", [True, False])
def test_scaled_gemm(
    m_dim: int,
    k_dim: int,
    n_dim: int,
    input_dtype: torch.dtype,
    output_dtype: torch.dtype,
    use_scalar_scale_a: bool,
    use_scalar_scale_b: bool,
    use_bias: bool,
) -> None:
    """Test mixed precision GEMM Triton kernel."""
    seed: Final = 0
    seed_everything(seed)

    device: Final = torch.device(current_platform.device)
    torch.set_default_device(device)

    if use_scalar_scale_a:
        scale_a_tensor = torch.rand((1, 1), dtype=torch.float32, device=device)
    else:
        scale_a_tensor = 0.25 * torch.rand((m_dim, 1), device=device)

    if use_scalar_scale_b:
        scale_b_tensor = torch.rand((1, 1), dtype=torch.float32, device=device)
    else:
        scale_b_tensor = 0.25 * torch.rand((n_dim, 1), device=device)

    if _is_floating_point_type(input_dtype):
        a = (0.25 * torch.rand((m_dim, k_dim), dtype=torch.float32, device=device)).to(input_dtype)
        b = (0.25 * torch.rand((n_dim, k_dim), dtype=torch.float32, device=device)).to(input_dtype).T
    else:
        a = torch.randint(-32, 32, (m_dim, k_dim), dtype=input_dtype, device=device)
        b = torch.randint(-32, 32, (n_dim, k_dim), dtype=input_dtype, device=device).T

    bias = None
    if use_bias:
        bias = torch.rand((n_dim,), device=device, dtype=output_dtype)

    reference_output = scaled_gemm_reference(a, b, scale_a_tensor, scale_b_tensor, output_dtype, bias)
    triton_output = scaled_gemm_triton(a, b, scale_a_tensor, scale_b_tensor, output_dtype, bias)

    torch.testing.assert_close(reference_output, triton_output, rtol=1e-1, atol=1e-1)



# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

"""Reference implementation of Gemma RMS norm kernel."""

import torch

from conch import envs
from conch.platforms import current_platform


def _gemma_rms_norm_pytorch_ref(
    weight: torch.Tensor,
    variance_epsilon: float,
    x: torch.Tensor,
    residual: torch.Tensor | None,
) -> torch.Tensor | tuple[torch.Tensor, torch.Tensor]:
    """PyTorch reference gemma_rms_norm impl."""
    orig_dtype = x.dtype
    if residual is not None:
        x = x + residual
        residual = x

    x = x.float()
    variance = x.pow(2).mean(dim=-1, keepdim=True)
    x = x * torch.rsqrt(variance + variance_epsilon)
    # Llama does x.to(float16) * w whilst Gemma is (x * w).to(float16)
    # See https://github.com/huggingface/transformers/pull/29402
    x = x * (1.0 + weight.float())
    x = x.to(orig_dtype)
    return x if residual is None else (x, residual)


def _gemma_rms_norm_vllm_ref(
    weight: torch.Tensor,
    variance_epsilon: float,
    x: torch.Tensor,
    residual: torch.Tensor | None,
) -> torch.Tensor | tuple[torch.Tensor, torch.Tensor]:
    """vLLM reference gemma_rms_norm impl."""
    from vllm.model_executor.layers.layernorm import GemmaRMSNorm

    layer = GemmaRMSNorm(hidden_size=weight.size(0), eps=variance_epsilon)
    layer.weight = torch.nn.Parameter(weight)

    return layer.forward_cuda(x, residual)  # type: ignore[no-any-return, unused-ignore]


def gemma_rms_norm(
    x: torch.Tensor,
    weight: torch.Tensor,
    variance_epsilon: float,
    residual: torch.Tensor | None = None,
) -> torch.Tensor | tuple[torch.Tensor, torch.Tensor]:
    """Gemma RMS norm operation."""
    if envs.CONCH_ENABLE_VLLM and current_platform.has_cuda():
        return _gemma_rms_norm_vllm_ref(
            weight,
            variance_epsilon,
            x,
            residual,
        )

    return _gemma_rms_norm_pytorch_ref(
        weight,
        variance_epsilon,
        x,
        residual,
    )



# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

"""Test Triton varlen attention."""

from typing import Final

import pytest
import torch

from conch import envs
from conch.ops.attention.varlen_attention import varlen_attention
from conch.platforms import current_platform
from conch.third_party.vllm.utils import create_tensors, seed_everything

_ENABLE_VLLM: Final = envs.CONCH_ENABLE_VLLM and current_platform.has_cuda()
_HEAD_SIZES: Final = [64, 96, 128, 256]
_NUM_SEQS_ABRIDGED: Final = [4, 9]
# MHA, MQA, and GQA
# - MHA: num_query_heads == num_kv_heads
# - MQA: num_kv_heads == 1
# - GQA: num_query_heads != num_kv_heads && num_kv_heads != 1
_NUM_HEADS_ABRIDGED: Final = [(8, 8), (4, 1), (16, 4)]
_SEQUENCE_LENGTHS: Final = [256, 257, 343, 1024, 1025]


def _get_tolerance_for_dtype(dtype: torch.dtype) -> float:
    """Get expected tolerance to match to for a given dtype."""
    if dtype == torch.float16:
        return 7e-4

    if dtype == torch.bfloat16:
        return 2e-3

    msg = f"Unsupported dtype: '{dtype}'"
    raise NotImplementedError(msg)


def _convert_paged_to_contiguous(
    k_cache_paged: torch.Tensor,
    v_cache_paged: torch.Tensor,
    block_table: torch.Tensor,
    cu_seqlens_k: torch.Tensor,
    total_num_k: int,
    num_kv_heads: int,
    head_size: int,
    cache_block_size: int,
) -> tuple[torch.Tensor, torch.Tensor]:
    """Convert paged kv cache to contiguous.

    Args:
        k_cache_paged: Paged key cache.
        v_cache_paged: Paged value cache.
        block_table: Block tables for traversing kv caches.
        cu_seqlens_k: Sequence lengths.
        num_kv_heads: The number of kv heads.
        total_num_k: The total number of keys/values.
        head_size: The dimension of the attention head.
        cache_block_size: The size of each cache block.

    Returns:
        Tuple of key cache in contiguous memory and value cache in contiguous memory.
    """
    batch_size, _ = block_table.shape

    k_contiguous = torch.empty(
        (total_num_k, num_kv_heads, head_size), dtype=k_cache_paged.dtype, device=k_cache_paged.device
    )
    v_contiguous = torch.empty(
        (total_num_k, num_kv_heads, head_size), dtype=v_cache_paged.dtype, device=v_cache_paged.device
    )

    for batch_index in range(batch_size):
        current_seq_len_begin = int(cu_seqlens_k[batch_index].item())
        current_seq_len_end = int(cu_seqlens_k[batch_index + 1].item())
        this_seqlen = current_seq_len_end - current_seq_len_begin

        this_k = k_contiguous[current_seq_len_begin:current_seq_len_end]
        this_v = v_contiguous[current_seq_len_begin:current_seq_len_end]

        these_blocks = block_table[batch_index]

        relative_begin = 0

        for relative_block_index, cache_block_index in enumerate(these_blocks):
            current_k_block = k_cache_paged[cache_block_index]
            current_v_block = v_cache_paged[cache_block_index]

            relative_seqlen = relative_block_index * cache_block_size
            if relative_seqlen >= this_seqlen:
                break

            # Relative end is the end index in the current cache block
            relative_end = relative_begin + min(this_seqlen - relative_seqlen, cache_block_size)

            this_length = relative_end - relative_begin

            this_k[relative_begin:relative_end, :, :] = current_k_block[:this_length, :, :]
            this_v[relative_begin:relative_end, :, :] = current_v_block[:this_length, :, :]

            relative_begin += cache_block_size

    return k_contiguous, v_contiguous


def _pytorch_attention_inner(
    q: torch.Tensor,
    k: torch.Tensor,
    v: torch.Tensor,
    scale: float,
    causal: bool,
) -> tuple[torch.Tensor, torch.Tensor]:
    attention_scores = torch.matmul(q, k.transpose(-2, -1))

    attention_scaled_scores = scale * attention_scores

    if causal:
        q_len, k_len = q.shape[1], k.shape[1]
        row_idx = torch.arange(q_len, device=q.device).unsqueeze(1)
        col_idx = torch.arange(k_len, device=q.device).unsqueeze(0)
        col_offset = q_len - k_len
        causal_mask = row_idx >= (col_offset + col_idx)
        attention_scaled_scores = attention_scaled_scores.masked_fill(
            torch.logical_not(causal_mask.unsqueeze(0)), float("-inf")
        )

    max_scores = torch.max(attention_scaled_scores, dim=-1, keepdim=True)[0]
    if causal:
        max_scores = torch.where(torch.isinf(max_scores), torch.zeros_like(max_scores), max_scores)

    attention_shifted_scaled_scores = attention_scaled_scores - max_scores

    exp_scores = torch.exp(attention_shifted_scaled_scores)

    sum_exp_scores = torch.sum(exp_scores, dim=-1, keepdim=True)
    if causal:
        sum_exp_scores = torch.where(sum_exp_scores == 0, torch.ones_like(sum_exp_scores), sum_exp_scores)

    p = exp_scores / sum_exp_scores

    softmax_lse = max_scores + torch.log(sum_exp_scores)
    softmax_lse = softmax_lse.squeeze(-1)

    output = torch.matmul(p, v)

    return output, softmax_lse


def _varlen_attention_pytorch(
    q: torch.Tensor,
    k: torch.Tensor,
    v: torch.Tensor,
    cu_seqlens_q: torch.Tensor,
    cu_seqlens_k: torch.Tensor,
    max_seqlen_q: int,
    max_seqlen_k: int,
    scale: float,
    causal: bool,
) -> torch.Tensor:
    batch_size = cu_seqlens_q.shape[0] - 1
    num_query_heads, num_kv_heads = q.shape[1], k.shape[1]
    head_size = q.shape[2]

    total_num_q = q.shape[0]

    output = torch.zeros((total_num_q, num_query_heads, head_size), dtype=q.dtype, device=q.device)

    group_size = num_query_heads // num_kv_heads
    if num_query_heads % num_kv_heads != 0:
        raise ValueError("num_query_heads must be divisible by num_kv_heads")

    for batch_index in range(batch_size):
        start_q = int(cu_seqlens_q[batch_index].item())
        end_q = int(cu_seqlens_q[batch_index + 1].item())
        start_k = int(cu_seqlens_k[batch_index].item())
        end_k = int(cu_seqlens_k[batch_index + 1].item())

        seqlen_q = end_q - start_q
        seqlen_k = end_k - start_k

        this_q = q[start_q:end_q, :, :]
        this_k = k[start_k:end_k, :, :]
        this_v = v[start_k:end_k, :, :]

        this_q = this_q.permute(1, 0, 2)
        this_k = this_k.permute(1, 0, 2)
        this_v = this_v.permute(1, 0, 2)

        if group_size != 1:
            this_q = this_q.reshape(num_kv_heads, group_size, seqlen_q, head_size)
            this_k = this_k.unsqueeze(1).expand(-1, group_size, -1, -1)
            this_v = this_v.unsqueeze(1).expand(-1, group_size, -1, -1)

            this_q = this_q.reshape(num_kv_heads * group_size, seqlen_q, head_size)
            this_k = this_k.reshape(num_kv_heads * group_size, seqlen_k, head_size)
            this_v = this_v.reshape(num_kv_heads * group_size, seqlen_k, head_size)
        else:
            this_q = this_q.reshape(num_query_heads, seqlen_q, head_size)
            this_k = this_k.reshape(num_kv_heads, seqlen_k, head_size)
            this_v = this_v.reshape(num_kv_heads, seqlen_k, head_size)

        this_output, _ = _pytorch_attention_inner(this_q, this_k, this_v, scale, causal)

        if group_size != 1:
            this_output = this_output.reshape(num_kv_heads, group_size, seqlen_q, head_size)
            this_output = this_output.reshape(num_query_heads, seqlen_q, head_size)

        this_output = this_output.permute(1, 0, 2)

        output[start_q:end_q, :, :] = this_output

    return output


@pytest.mark.parametrize("num_seqs", _NUM_SEQS_ABRIDGED)
@pytest.mark.parametrize("head_size", _HEAD_SIZES)
@pytest.mark.parametrize(("num_query_heads", "num_kv_heads"), _NUM_HEADS_ABRIDGED)
@pytest.mark.parametrize("dtype", [torch.float16, torch.bfloat16])
@pytest.mark.parametrize("sequence_length", _SEQUENCE_LENGTHS)
@pytest.mark.parametrize("causal", [True, False])
def test_varlen_attention_vs_pytorch(
    num_seqs: int,
    head_size: int,
    num_query_heads: int,
    num_kv_heads: int,
    dtype: torch.dtype,
    sequence_length: int,
    causal: bool,
) -> None:
    """Test Varlen Attention Triton kernel with various configurations vs. vLLM FlashAttnVarlen."""
    seed: Final = 0
    seed_everything(seed)

    device: Final = torch.device(current_platform.device)
    torch.set_default_device(device)

    scale: Final = float(1.0 / (head_size**0.5))
    softcap: Final = 0.0

    tolerance: Final = _get_tolerance_for_dtype(dtype)

    kv_cache_dtype: Final = "auto"

    cache_block_size = 16

    _, _, _, key_cache, value_cache, block_table, seq_lens = create_tensors(
        head_size,
        sequence_length,
        cache_block_size,
        num_seqs,
        num_query_heads,
        num_kv_heads,
        kv_cache_dtype,
        current_platform.device,
        dtype,
    )

    starting_item = torch.as_tensor([0], dtype=torch.int32)

    cu_seqlens_q = torch.cumsum(seq_lens, dim=0, dtype=torch.int32)
    cu_seqlens_q = torch.cat((starting_item, cu_seqlens_q), dim=0)

    cu_seqlens_k = cu_seqlens_q.clone()

    max_seqlen_q = int(torch.max(seq_lens).item())
    max_seqlen_k = int(max_seqlen_q)

    total_num_q = int(cu_seqlens_q[-1].item())
    total_num_k = int(cu_seqlens_k[-1].item())

    key_contiguous, value_contiguous = _convert_paged_to_contiguous(
        key_cache,
        value_cache,
        block_table,
        cu_seqlens_k,
        total_num_k,
        num_kv_heads,
        head_size,
        cache_block_size,
    )

    q = torch.empty(total_num_q, num_query_heads, head_size, dtype=dtype, device=device)
    q.uniform_(-scale, scale)

    pytorch_output = _varlen_attention_pytorch(
        q=q,
        k=key_contiguous,
        v=value_contiguous,
        cu_seqlens_q=cu_seqlens_q,
        cu_seqlens_k=cu_seqlens_k,
        max_seqlen_q=max_seqlen_q,
        max_seqlen_k=max_seqlen_k,
        scale=scale,
        causal=causal,
    )

    conch_output = varlen_attention(
        query=q,
        key_cache=key_cache,
        value_cache=value_cache,
        cu_seqlens_q=cu_seqlens_q,
        max_seqlen_q=max_seqlen_q,
        seq_lens=seq_lens,
        max_seqlen_k=max_seqlen_k,
        block_table=block_table,
        causal=causal,
        scale=scale,
        softcap=softcap,
    )

    torch.testing.assert_close(pytorch_output, conch_output, atol=tolerance, rtol=tolerance)


@pytest.mark.skipif(not _ENABLE_VLLM, reason="This test case requires vLLM")
@pytest.mark.parametrize("num_seqs", _NUM_SEQS_ABRIDGED)
@pytest.mark.parametrize("head_size", _HEAD_SIZES)
@pytest.mark.parametrize(("num_query_heads", "num_kv_heads"), _NUM_HEADS_ABRIDGED)
@pytest.mark.parametrize("dtype", [torch.float16, torch.bfloat16])
@pytest.mark.parametrize("sequence_length", _SEQUENCE_LENGTHS)
@pytest.mark.parametrize("causal", [True, False])
@pytest.mark.parametrize("is_pure_decode", [True, False])
@pytest.mark.parametrize("kv_cache_dtype", ["auto", "fp8"])
def test_varlen_attention_vs_flash_attn(
    num_seqs: int,
    head_size: int,
    num_query_heads: int,
    num_kv_heads: int,
    dtype: torch.dtype,
    sequence_length: int,
    causal: bool,
    is_pure_decode: bool,
    kv_cache_dtype: str,
) -> None:
    """Test Varlen Attention Triton kernel with various configurations vs. vLLM FlashAttnVarlen."""
    if kv_cache_dtype == "fp8" and not current_platform.supports_fp8():
        pytest.skip()

    from vllm.vllm_flash_attn import flash_attn_varlen_func  # type: ignore[attr-defined, unused-ignore]

    seed: Final = 0
    seed_everything(seed)

    device: Final = torch.device(current_platform.device)
    torch.set_default_device(device)

    scale: Final = float(1.0 / (head_size**0.5))
    softcap: Final = 0.0

    tolerance: Final = _get_tolerance_for_dtype(dtype)

    cache_block_size = 16

    _, _, _, key_cache, value_cache, block_table, seq_lens = create_tensors(
        head_size,
        sequence_length,
        cache_block_size,
        num_seqs,
        num_query_heads,
        num_kv_heads,
        kv_cache_dtype,
        current_platform.device,
        dtype,
    )

    q_scale = torch.full((1,), 0.5, device=device)
    k_scale = torch.full((1,), 0.5, device=device)
    v_scale = torch.full((1,), 0.5, device=device)

    starting_item = torch.as_tensor([0], dtype=torch.int32)

    if is_pure_decode:
        seqlens_q = torch.ones((num_seqs,), dtype=torch.int32)

        cu_seqlens_q = torch.cumsum(seqlens_q, dim=0, dtype=torch.int32)
        cu_seqlens_q = torch.cat((starting_item, cu_seqlens_q), dim=0)

        max_seqlen_q = int(torch.max(seqlens_q).item())
        max_seqlen_k = int(torch.max(seq_lens).item())
    else:
        cu_seqlens_q = torch.cumsum(seq_lens, dim=0, dtype=torch.int32)
        cu_seqlens_q = torch.cat((starting_item, cu_seqlens_q), dim=0)

        max_seqlen_q = int(torch.max(seq_lens).item())
        max_seqlen_k = int(max_seqlen_q)

    batch_size = cu_seqlens_q.shape[0] - 1
    total_num_q = int(cu_seqlens_q[-1].item())

    q = torch.empty(total_num_q, num_query_heads, head_size, dtype=dtype, device=device)
    q.uniform_(-scale, scale)

    # FA version 2 is fine until we need FP8 support
    fa_version = 2

    if kv_cache_dtype == "fp8":
        fp8_dtype = torch.float8_e4m3fnuz if current_platform.is_amd() else torch.float8_e4m3fn
        q = q.to(fp8_dtype)
        key_cache = key_cache.view(fp8_dtype)
        value_cache = value_cache.view(fp8_dtype)
        # Use FA version 3 if we are using FP8
        fa_version = 3

    vllm_output = flash_attn_varlen_func(
        q=q,
        k=key_cache,
        v=value_cache,
        cu_seqlens_q=cu_seqlens_q,
        max_seqlen_q=max_seqlen_q,
        max_seqlen_k=max_seqlen_k,
        block_table=block_table,
        seqused_k=seq_lens,
        softmax_scale=scale,
        causal=causal,
        fa_version=fa_version,
        q_descale=q_scale.expand(batch_size, num_kv_heads),
        k_descale=k_scale.expand(batch_size, num_kv_heads),
        v_descale=v_scale.expand(batch_size, num_kv_heads),
    )

    conch_output = varlen_attention(
        query=q,
        key_cache=key_cache,
        value_cache=value_cache,
        cu_seqlens_q=cu_seqlens_q,
        max_seqlen_q=max_seqlen_q,
        seq_lens=seq_lens,
        max_seqlen_k=max_seqlen_k,
        block_table=block_table,
        causal=causal,
        scale=scale,
        softcap=softcap,
        kv_cache_dtype=kv_cache_dtype,
        q_scale=q_scale,
        k_scale=k_scale,
        v_scale=v_scale,
    )

    torch.testing.assert_close(vllm_output, conch_output, atol=tolerance, rtol=tolerance)


@pytest.mark.skipif(not _ENABLE_VLLM, reason="This test case requires vLLM")
@pytest.mark.parametrize("num_seqs", _NUM_SEQS_ABRIDGED)
@pytest.mark.parametrize("head_size", _HEAD_SIZES)
@pytest.mark.parametrize(("num_query_heads", "num_kv_heads"), _NUM_HEADS_ABRIDGED)
@pytest.mark.parametrize("dtype", [torch.float16, torch.bfloat16])
@pytest.mark.parametrize("sequence_length", [8192])
@pytest.mark.parametrize("causal", [True])
@pytest.mark.parametrize("kv_cache_dtype", ["auto", "fp8"])
def test_varlen_attention_decode(
    num_seqs: int,
    head_size: int,
    num_query_heads: int,
    num_kv_heads: int,
    dtype: torch.dtype,
    sequence_length: int,
    causal: bool,
    kv_cache_dtype: str,
) -> None:
    """Test Varlen Attention Triton kernel with various configurations vs. vLLM FlashAttnVarlen."""
    if kv_cache_dtype == "fp8" and not current_platform.supports_fp8():
        pytest.skip()

    from vllm.vllm_flash_attn import flash_attn_varlen_func  # type: ignore[attr-defined, unused-ignore]

    seed: Final = 0
    seed_everything(seed)

    device: Final = torch.device(current_platform.device)
    torch.set_default_device(device)

    scale: Final = float(1.0 / (head_size**0.5))
    softcap: Final = 0.0

    tolerance: Final = _get_tolerance_for_dtype(dtype)

    cache_block_size = 16

    _, _, _, key_cache, value_cache, block_table, seq_lens = create_tensors(
        head_size,
        sequence_length,
        cache_block_size,
        num_seqs,
        num_query_heads,
        num_kv_heads,
        kv_cache_dtype,
        current_platform.device,
        dtype,
    )

    q_scale = torch.full((1,), 0.5, device=device)
    k_scale = torch.full((1,), 0.5, device=device)
    v_scale = torch.full((1,), 0.5, device=device)

    starting_item = torch.as_tensor([0], dtype=torch.int32)

    # All seqlen_q == 1 because its all decodes
    seqlens_q = torch.ones((num_seqs,), dtype=torch.int32)

    cu_seqlens_q = torch.cumsum(seqlens_q, dim=0, dtype=torch.int32)
    cu_seqlens_q = torch.cat((starting_item, cu_seqlens_q), dim=0)

    max_seqlen_q = int(torch.max(seqlens_q).item())
    max_seqlen_k = int(torch.max(seq_lens).item())

    batch_size = cu_seqlens_q.shape[0] - 1
    total_num_q = int(cu_seqlens_q[-1].item())

    q = torch.empty(total_num_q, num_query_heads, head_size, dtype=dtype, device=device)
    q.uniform_(-scale, scale)

    # FA version 2 is fine until we need FP8 support
    fa_version = 2

    if kv_cache_dtype == "fp8":
        fp8_dtype = torch.float8_e4m3fnuz if current_platform.is_amd() else torch.float8_e4m3fn
        q = q.to(fp8_dtype)
        key_cache = key_cache.view(fp8_dtype)
        value_cache = value_cache.view(fp8_dtype)
        # Use FA version 3 if we are using FP8
        fa_version = 3

    vllm_output = flash_attn_varlen_func(
        q=q,
        k=key_cache,
        v=value_cache,
        cu_seqlens_q=cu_seqlens_q,
        max_seqlen_q=max_seqlen_q,
        max_seqlen_k=max_seqlen_k,
        block_table=block_table,
        seqused_k=seq_lens,
        softmax_scale=scale,
        causal=causal,
        fa_version=fa_version,
        q_descale=q_scale.expand(batch_size, num_kv_heads),
        k_descale=k_scale.expand(batch_size, num_kv_heads),
        v_descale=v_scale.expand(batch_size, num_kv_heads),
    )

    conch_output = varlen_attention(
        query=q,
        key_cache=key_cache,
        value_cache=value_cache,
        cu_seqlens_q=cu_seqlens_q,
        max_seqlen_q=max_seqlen_q,
        seq_lens=seq_lens,
        max_seqlen_k=max_seqlen_k,
        block_table=block_table,
        causal=causal,
        scale=scale,
        softcap=softcap,
        kv_cache_dtype=kv_cache_dtype,
        q_scale=q_scale,
        k_scale=k_scale,
        v_scale=v_scale,
    )

    torch.testing.assert_close(vllm_output, conch_output, atol=tolerance, rtol=tolerance)


@pytest.mark.skipif(not _ENABLE_VLLM, reason="This test case requires vLLM")
@pytest.mark.parametrize("dtype", [torch.float16, torch.bfloat16])
def test_vllm_crash(dtype: torch.dtype) -> None:
    """Test Varlen Attention Triton kernel with various configurations vs. vLLM FlashAttnVarlen."""
    from vllm.vllm_flash_attn import flash_attn_varlen_func  # type: ignore[attr-defined, unused-ignore]

    head_size: Final = 128
    num_query_heads: Final = 32
    num_kv_heads: Final = 8
    cache_block_size: Final = 128

    seed: Final = 0
    seed_everything(seed)

    device: Final = torch.device(current_platform.device)
    torch.set_default_device(device)

    scale: Final = float(1.0 / (head_size**0.5))
    softcap: Final = 0.0

    tolerance = 1e-2 if dtype == torch.bfloat16 else 1e-3

    block_table = torch.tensor(
        [
            [1, 2, 3, 4, 10, 6, 7, 8, 9, 0, 0, 0, 0, 0, 0, 0],
            [11, 12, 13, 14, 15, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [16, 17, 18, 19, 20, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [21, 22, 23, 24, 25, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [26, 27, 28, 29, 30, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [31, 32, 33, 34, 35, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [36, 37, 38, 39, 40, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [41, 42, 43, 44, 45, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [46, 47, 48, 49, 50, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [51, 52, 53, 54, 55, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [56, 57, 58, 59, 60, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [61, 62, 63, 64, 65, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [66, 67, 68, 69, 70, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [71, 72, 73, 74, 75, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [76, 77, 78, 79, 80, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [81, 82, 83, 84, 85, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [86, 87, 88, 89, 90, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [91, 92, 93, 94, 95, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [96, 97, 98, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        ],
        device=device,
        dtype=torch.int32,
    )

    cu_seqlens_q = torch.tensor(
        [0, 1, 2, 3, 534, 1054, 1580, 2095, 2619, 3144, 3681, 4211, 4743, 5261, 5794, 6327, 6842, 7365, 7890, 8192],
        device=device,
        dtype=torch.int32,
    )
    seq_lens = torch.tensor(
        [536, 530, 520, 531, 520, 526, 515, 524, 525, 537, 530, 532, 518, 533, 533, 515, 523, 525, 302],
        device=device,
        dtype=torch.int32,
    )

    key_cache = torch.randn(3236, cache_block_size, num_kv_heads, head_size, dtype=dtype, device=device)
    value_cache = torch.randn(3236, cache_block_size, num_kv_heads, head_size, dtype=dtype, device=device)

    q = torch.empty(8192, num_query_heads, head_size, dtype=dtype, device=device)
    q.uniform_(-scale, scale)

    max_seqlen_q = int(torch.max(seq_lens).item())
    max_seqlen_k = max_seqlen_q

    causal = True

    vllm_output = flash_attn_varlen_func(
        q=q,
        k=key_cache,
        v=value_cache,
        cu_seqlens_q=cu_seqlens_q,
        max_seqlen_q=max_seqlen_q,
        max_seqlen_k=max_seqlen_k,
        block_table=block_table,
        seqused_k=seq_lens,
        softmax_scale=scale,
        causal=causal,
    )

    conch_output = varlen_attention(
        query=q,
        key_cache=key_cache,
        value_cache=value_cache,
        cu_seqlens_q=cu_seqlens_q,
        max_seqlen_q=max_seqlen_q,
        seq_lens=seq_lens,
        max_seqlen_k=max_seqlen_k,
        block_table=block_table,
        causal=causal,
        scale=scale,
        softcap=softcap,
    )

    torch.testing.assert_close(vllm_output, conch_output, atol=tolerance, rtol=tolerance)



# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

"""Conch paged attention vs. FlashAttnWithKVCache benchmark."""

import sys
from typing import Final

import click
import torch

from conch import envs
from conch.ops.attention.paged_attention import paged_attention
from conch.platforms import current_platform
from conch.third_party.vllm.utils import create_tensors, seed_everything
from conch.utils.benchmark import BenchmarkMetadata, benchmark_it

if envs.CONCH_ENABLE_VLLM and current_platform.is_nvidia():
    from vllm.vllm_flash_attn import flash_attn_with_kvcache  # type: ignore[attr-defined, unused-ignore]
else:
    flash_attn_with_kvcache = None  # type: ignore[assignment, unused-ignore]


@click.command()
@click.option(
    "--head-dim",
    required=True,
    type=int,
    default=128,
    help="Head dimension",
)
@click.option(
    "--seq-len",
    required=True,
    type=int,
    default=8192,
    help="Sequence length (for k/v)",
)
@click.option(
    "--cache-block-size",
    required=True,
    type=int,
    default=32,
    help="Number of KV vectors in each cache block",
)
@click.option(
    "--batch-size",
    required=False,
    type=int,
    default=128,
    help="Batch size",
)
@click.option(
    "--num-query-heads",
    required=False,
    type=int,
    default=32,
    help="Number of query heads",
)
@click.option(
    "--num-kv-heads",
    required=False,
    type=int,
    default=8,
    help="Number of kv heads",
)
@click.option(
    "--iteration-time-ms",
    required=False,
    type=int,
    default=10000,
    help="Time in milliseconds to run benchmark",
)
@click.option(
    "--warmup-time-ms",
    required=False,
    type=int,
    default=1000,
    help="Time in milliseconds to warmup before recording times",
)
@click.option(
    "--absolute-tolerance",
    required=False,
    type=float,
    default=1e-3,
    help="Absolute tolerance to match with",
)
@click.option(
    "--verbose",
    is_flag=True,
    help="Flag for printing verbose output",
)
@click.option(
    "--gpu",
    required=False,
    type=str,
    default=current_platform.device,
    help="Device to run on",
)
@click.option(
    "--csv",
    is_flag=True,
    help="Flag for printing results in CSV format",
)
def main(
    head_dim: int,
    seq_len: int,
    cache_block_size: int,
    batch_size: int,
    num_query_heads: int,
    num_kv_heads: int,
    iteration_time_ms: int,
    warmup_time_ms: int,
    absolute_tolerance: float,
    verbose: bool,
    gpu: str,
    csv: bool,
) -> None:
    """Benchmark PagedAttention.

    Args:
        head_dim: Head dimension of input tensors.
        seq_len: Sequence length of input tensors.
        cache_block_size: Number of KV vectors in each cache block.
        batch_size: Batch size of input tensors.
        num_query_heads: Number of attention query heads.
        num_kv_heads: Number of attention kv heads.
        iteration_time_ms: Time in milliseconds to run benchmark.
        warmup_time_ms: Time in milliseconds to warmup before recording times.
        absolute_tolerance: Absolute tolerance used to check accuracy of PyTorch vs. Conch.
        verbose: Flag to indicate whether or not to print verbose output.
        gpu: Which gpu to run on.
        csv: Flag to indicate whether or not to print results in CSV format.
    """
    if not current_platform.is_nvidia() or flash_attn_with_kvcache is None:
        error_msg = "Platform must be Nvidia and vLLM must be installed & enabled via CONCH_ENABLE_VLLM=1"
        raise NotImplementedError(error_msg)

    seed: Final = 0
    seed_everything(seed)

    device: Final = torch.device(gpu)
    torch.set_default_device(device)

    dtype: Final = torch.float16

    metadata = BenchmarkMetadata(
        platform=current_platform.name(),
        params={
            "head_dim": head_dim,
            "seq_len": seq_len,
            "cache_block_size": cache_block_size,
            "batch_size": batch_size,
            "num_query_heads": num_query_heads,
            "num_kv_heads": num_kv_heads,
        },
    )

    kv_cache_dtype = "auto"

    query, _, _, key_cache, value_cache, block_table, seq_lens = create_tensors(
        head_dim, seq_len, cache_block_size, batch_size, num_query_heads, num_kv_heads, kv_cache_dtype, device, dtype
    )

    _, max_num_blocks_per_seq = block_table.shape

    scale: Final = float(1.0 / (head_dim**0.5))

    alibi_slopes = None

    # Create output tensors
    query_vllm = query.unsqueeze(1)
    output_conch = torch.empty_like(query)

    softcap = 30.0
    k_scale = torch.full((1,), 1.0, dtype=dtype, device=device)
    v_scale = torch.full((1,), 1.0, dtype=dtype, device=device)

    # Check accuracy match
    output_vllm = flash_attn_with_kvcache(
        query_vllm,
        key_cache,
        value_cache,
        block_table=block_table,
        cache_seqlens=seq_lens,
        softmax_scale=scale,
        causal=True,
        alibi_slopes=alibi_slopes,
        softcap=softcap,
    )

    output_vllm = output_vllm.squeeze(1)

    paged_attention(
        query,
        key_cache,
        value_cache,
        block_table,
        seq_lens,
        output=output_conch,
        scale=scale,
        softcap=softcap,
        kv_cache_dtype=kv_cache_dtype,
        k_scale=k_scale,
        v_scale=v_scale,
    )

    if not torch.allclose(output_vllm, output_conch, atol=absolute_tolerance):
        print(f"WARNING: Reference and Conch results differ! (atol={absolute_tolerance})", file=sys.stderr)
        print(f"Output max diff: {(output_conch - output_vllm).abs().max().item()}", file=sys.stderr)

        if verbose:
            print(f"Reference output: {output_vllm}", file=sys.stderr)
            print(f"Conch output: {output_conch}", file=sys.stderr)
    else:
        print(f"Results matched with atol={absolute_tolerance} :)", file=sys.stderr)

    baseline_result = benchmark_it(
        lambda: flash_attn_with_kvcache(
            query_vllm,
            key_cache,
            value_cache,
            block_table=block_table,
            cache_seqlens=seq_lens,
            softmax_scale=scale,
            causal=True,
            alibi_slopes=alibi_slopes,
        ),
        tag="Baseline",
        metadata=metadata,
    )

    conch_result = benchmark_it(
        lambda: paged_attention(
            query,
            key_cache,
            value_cache,
            block_table,
            seq_lens,
            output=output_conch,
            scale=scale,
            softcap=softcap,
            kv_cache_dtype=kv_cache_dtype,
            k_scale=k_scale,
            v_scale=v_scale,
        ),
        tag="Conch",
        metadata=metadata,
    )

    # Print results
    conch_result.print_parameters(csv=csv)
    conch_result.print_results(csv=csv)
    baseline_result.print_results(csv=csv)


if __name__ == "__main__":
    main()



# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

"""Utilities for benchmarking."""

from collections.abc import Callable
from dataclasses import dataclass
from typing import Any

import triton


@dataclass
class BenchmarkMetadata:
    """Class holding metadata for benchmark."""

    platform: str
    params: dict[str, Any]


@dataclass
class BenchmarkResult:
    """Class holding results of benchmark."""

    tag: str
    metadata: BenchmarkMetadata
    num_iterations: int
    min_: float
    max_: float
    mean_: float
    median_: float

    def print_parameters(self, csv: bool = False) -> None:
        """Print the parameters of the benchmark.

        Args:
            csv: (Optional) If True, print in CSV format.
        """
        if csv:
            print(f"tag,platform,num_iterations,{','.join(self.metadata.params.keys())},runtime_ms")
            return

        print(f"Parameters: {self.metadata.params}")

    def csv_print(self) -> None:
        """Convert the benchmark result to a CSV string.

        Returns:
            A CSV string representation of the benchmark result.
        """
        print(
            f"{self.tag},{self.metadata.platform},{self.num_iterations},{','.join(str(v) for v in self.metadata.params.values())},{self.median_:.3f}"
        )

    def pretty_print(self) -> None:
        """Pretty print the benchmarking results.

        Args:
            unit: (Optional) Unit to print the times in, default is milliseconds.
        """

        def _format(value: float) -> str:
            return f"{value:.3f} ms"

        print(
            f"{self.tag}: num_iterations={self.num_iterations}, min={_format(self.min_)}, max={_format(self.max_)}, mean={_format(self.mean_)}, median={_format(self.median_)}"
        )

    def print_results(self, csv: bool = False) -> None:
        """Print the benchmark results.

        Args:
            csv: (Optional) If True, print in CSV format.
        """
        if csv:
            self.csv_print()
            return

        self.pretty_print()


def benchmark_it(
    fn: Callable[[], Any],
    tag: str,
    metadata: BenchmarkMetadata,
    iteration_time_ms: int = 10000,
    warmup_time_ms: int = 1000,
) -> BenchmarkResult:
    """Function to benchmark a function.

    Args:
        tag: The tag to identify the benchmark.
        fn: The function to benchmark.
        metadata: Metadata for the benchmark.
        iteration_time_ms: Time in milliseconds to run the benchmark.
        warmup_time_ms: Time in milliseconds to warm up before recording times.

    Returns:
        BenchmarkResult dataclass with aggregated results of benchmark.
    """
    results = triton.testing.do_bench(fn, warmup=warmup_time_ms, rep=iteration_time_ms, return_mode="all")
    num_iterations = len(results)

    return BenchmarkResult(
        tag,
        metadata,
        num_iterations,
        min(results),
        max(results),
        sum(results) / num_iterations,
        sorted(results)[num_iterations // 2],
    )



# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

"""Scaled matrix multiplication kernel benchmark."""

import sys
from typing import Final

import click
import torch

from conch.ops.quantization.gemm import scaled_gemm as scaled_gemm_conch
from conch.platforms import current_platform
from conch.reference.quantization.scaled_gemm import scaled_gemm as scaled_gemm_reference
from conch.third_party.vllm.utils import seed_everything
from conch.utils.benchmark import BenchmarkMetadata, benchmark_it


def _to_torch_dtype(dtype_str: str) -> torch.dtype:
    """Map click arg for dtype to torch type."""
    if dtype_str == "bf16":
        return torch.bfloat16
    if dtype_str == "fp16":
        return torch.float16
    if dtype_str == "int8":
        return torch.int8
    if dtype_str == "fp8":
        return torch.float8_e4m3fnuz if current_platform.is_amd() else torch.float8_e4m3fn

    error_msg = f"Unrecognized data type: '{dtype_str}'"
    raise ValueError(error_msg)


def _is_floating_point_type(dtype: torch.dtype) -> bool:
    """Check whether a type is floating point."""
    return torch.tensor([1, 1], dtype=dtype).is_floating_point()


@click.command()
@click.option(
    "--m-dim",
    required=True,
    type=int,
    default=4096,
    help="1st dimension of A matrix (M x K)",
)
@click.option(
    "--k-dim",
    required=True,
    type=int,
    default=8192,
    help="Common dimension of A and B matrices (M x K) * (K * N)",
)
@click.option(
    "--n-dim",
    required=True,
    type=int,
    default=4096,
    help="2nd dimension of B matrix (K x N)",
)
@click.option(
    "--scale-a",
    required=True,
    type=float,
    default=2.1,
    help="Scaling arg for first input",
)
@click.option(
    "--scale-b",
    required=True,
    type=float,
    default=1.2,
    help="Scaling arg for second input",
)
@click.option(
    "--input-dtype",
    required=True,
    type=click.Choice(["fp8", "int8"]),
    default="int8",
    help="Data type of input",
)
@click.option(
    "--output-dtype",
    required=True,
    type=click.Choice(["fp16", "bf16"]),
    default="bf16",
    help="Data type of output",
)
@click.option(
    "--use-bias",
    is_flag=True,
    help="Flag for adding a bias",
)
@click.option(
    "--use-scalar-scale-a",
    is_flag=True,
    help="Flag for using scalar or vector for scale_a",
)
@click.option(
    "--use-scalar-scale-b",
    is_flag=True,
    help="Flag for using scalar or vector for scale_b",
)
@click.option(
    "--iteration-time-ms",
    required=False,
    type=int,
    default=10000,
    help="Time in milliseconds to run benchmark",
)
@click.option(
    "--warmup-time-ms",
    required=False,
    type=int,
    default=1000,
    help="Time in milliseconds to warmup before recording times",
)
@click.option(
    "--verbose",
    is_flag=True,
    help="Flag for printing verbose output",
)
@click.option(
    "--gpu",
    required=False,
    type=str,
    default=current_platform.device,
    help="Device to run on",
)
@click.option(
    "--csv",
    is_flag=True,
    help="Flag for printing results in CSV format",
)
def main(
    m_dim: int,
    k_dim: int,
    n_dim: int,
    scale_a: float,
    scale_b: float,
    input_dtype: str,
    output_dtype: str,
    use_bias: bool,
    use_scalar_scale_a: bool,
    use_scalar_scale_b: bool,
    iteration_time_ms: int,
    warmup_time_ms: int,
    verbose: bool,
    gpu: str,
    csv: bool,
) -> None:
    """Benchmark scaled matrix mulpity.

    Args:
        m_dim: 1st dimension of matrix A.
        k_dim: Common dimension between matrix A and B.
        n_dim: 2nd dimension of matrix B.
        scale_a: Scaling factor to apply to matrix A.
        scale_b: Scaling factor to apply to matrix B.
        input_dtype: Data type of input matrices.
        output_dtype: Data type of output matrices.
        use_bias: Whether or not to use bias.
        use_scalar_scale_a: Whether or not to use scalar scale_a.
        use_scalar_scale_b: Whether or not to use scalar scale_b.
        iteration_time_ms: Time in milliseconds to run benchmark.
        warmup_time_ms: Time in milliseconds to warmup before recording times.
        verbose: Flag to indicate whether or not to print verbose output.
        gpu: Which gpu to run on.
        csv: Flag for printing results in CSV format.
    """
    if input_dtype == "fp8" and not current_platform.supports_fp8():
        error_msg = "FP8 not supported on this GPU, cannot run benchmark!"
        raise NotImplementedError(error_msg)

    seed: Final = 0
    seed_everything(seed)

    device: Final = torch.device(gpu)
    torch.set_default_device(device)

    input_dtype_torch: Final = _to_torch_dtype(input_dtype)
    output_dtype_torch: Final = _to_torch_dtype(output_dtype)

    metadata = BenchmarkMetadata(
        platform=current_platform.name(),
        params={
            "m_dim": m_dim,
            "k_dim": k_dim,
            "n_dim": n_dim,
            "input_dtype": input_dtype,
            "output_dtype": output_dtype,
        },
    )

    if use_scalar_scale_a:
        scale_a_tensor = torch.tensor(scale_a, dtype=torch.float32, device=device)
    else:
        scale_a_tensor = 0.25 * torch.rand((m_dim, 1), device=device)

    if use_scalar_scale_b:
        scale_b_tensor = torch.tensor(scale_b, dtype=torch.float32, device=device)
    else:
        scale_b_tensor = 0.25 * torch.rand((n_dim, 1), device=device)

    if _is_floating_point_type(input_dtype_torch):
        a = (0.25 * torch.rand((m_dim, k_dim), dtype=torch.float32, device=device)).to(input_dtype_torch)
        b = (0.25 * torch.rand((n_dim, k_dim), dtype=torch.float32, device=device)).to(input_dtype_torch).T
    else:
        a = torch.randint(-32, 32, (m_dim, k_dim), dtype=input_dtype_torch, device=device)
        b = torch.randint(-32, 32, (n_dim, k_dim), dtype=input_dtype_torch, device=device).T

    bias = None
    if use_bias:
        bias = torch.rand((n_dim,), device=device, dtype=output_dtype_torch)

    reference_output = scaled_gemm_reference(a, b, scale_a_tensor, scale_b_tensor, output_dtype_torch, bias)
    conch_output = scaled_gemm_conch(a, b, scale_a_tensor, scale_b_tensor, output_dtype_torch, bias)

    if not torch.allclose(reference_output, conch_output, atol=1e-1, rtol=1e-1):
        print("WARNING: Reference and conch results differ!", file=sys.stderr)
        print(f"Output max diff: {(reference_output - conch_output).abs().max().item()}", file=sys.stderr)

        if verbose:
            print(f"Reference output: {reference_output}", file=sys.stderr)
            print(f"Conch output: {conch_output}", file=sys.stderr)
    else:
        print("Results matched :)", file=sys.stderr)

    baseline_result = benchmark_it(
        lambda: scaled_gemm_reference(
            a,
            b,
            scale_a_tensor,
            scale_b_tensor,
            output_dtype_torch,
            bias,
        ),
        tag="Baseline",
        metadata=metadata,
        iteration_time_ms=iteration_time_ms,
        warmup_time_ms=warmup_time_ms,
    )

    conch_result = benchmark_it(
        lambda: scaled_gemm_conch(
            a,
            b,
            scale_a_tensor,
            scale_b_tensor,
            output_dtype_torch,
            bias,
        ),
        tag="Conch",
        metadata=metadata,
        iteration_time_ms=iteration_time_ms,
        warmup_time_ms=warmup_time_ms,
    )

    # Print results
    conch_result.print_parameters(csv=csv)
    conch_result.print_results(csv=csv)
    baseline_result.print_results(csv=csv)


if __name__ == "__main__":
    main()



# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

"""Test for gelu_tanh_and_mul Triton port."""

from typing import Final

import pytest
import torch

from conch.ops.activation.gelu_tanh_and_mul import gelu_tanh_and_mul as gelu_tanh_and_mul_triton
from conch.platforms import current_platform
from conch.reference.activation.gelu_tanh_and_mul import gelu_tanh_and_mul as gelu_tanh_and_mul_reference
from conch.third_party.vllm.utils import seed_everything

_NUM_TOKENS: Final = [128, 2048]
_DIM: Final = [2048, 13824]


@pytest.mark.parametrize("n_tokens", _NUM_TOKENS)
@pytest.mark.parametrize("d", _DIM)
@torch.inference_mode()
def test_kernel(n_tokens: int, d: int) -> None:
    seed_everything(0)

    device: Final = torch.device(current_platform.device)
    torch.set_default_device(device)

    projections = torch.rand((n_tokens, d * 2), device=device)

    ref_output = gelu_tanh_and_mul_reference(projections)
    triton_output = gelu_tanh_and_mul_triton(projections)

    torch.testing.assert_close(triton_output, ref_output)



# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

"""Reference implementation of rotary embedding kernel."""

import torch

from conch import envs
from conch.platforms import current_platform


def _compute_inv_freq(base: float, rotary_dim: int) -> torch.Tensor:
    """Compute the inverse frequency."""
    return 1.0 / (base ** (torch.arange(0, rotary_dim, 2, dtype=torch.float) / rotary_dim))  # type: ignore[no-any-return]


def compute_cos_sin_cache(base: float, rotary_dim: int, max_position_embeddings: int) -> torch.Tensor:
    """Compute the cos and sin cache."""
    inv_freq = _compute_inv_freq(base, rotary_dim)
    t = torch.arange(max_position_embeddings, dtype=torch.float)

    freqs = torch.einsum("i,j -> ij", t, inv_freq)
    cos = freqs.cos()
    sin = freqs.sin()
    return torch.cat((cos, sin), dim=-1)


def _apply_rotary_emb(
    x: torch.Tensor,
    cos: torch.Tensor,
    sin: torch.Tensor,
    *,
    is_neox_style: bool = True,
) -> torch.Tensor:
    """
    Args:
        x: [num_tokens, num_heads, head_size]
        cos: [num_tokens, head_size // 2]
        sin: [num_tokens, head_size // 2]
        is_neox_style: Whether to use the Neox-style or GPT-J-style rotary positional embeddings.
    """
    cos = cos.unsqueeze(-2).to(x.dtype)
    sin = sin.unsqueeze(-2).to(x.dtype)
    if is_neox_style:
        x1, x2 = torch.chunk(x, 2, dim=-1)
    else:
        x1 = x[..., ::2]
        x2 = x[..., 1::2]
    o1 = x1 * cos - x2 * sin
    o2 = x2 * cos + x1 * sin
    if is_neox_style:
        return torch.cat((o1, o2), dim=-1)
    return torch.stack((o1, o2), dim=-1).flatten(-2)


def _rotary_embedding_pytorch_ref(
    positions: torch.Tensor,
    query: torch.Tensor,
    key: torch.Tensor,
    cos_sin_cache: torch.Tensor,
    rotary_dim: int,
    head_size: int,
    *,
    is_neox_style: bool = True,
    offsets: torch.Tensor | None = None,
) -> tuple[torch.Tensor, torch.Tensor]:
    """PyTorch reference rotary_embedding impl."""
    if offsets is not None:
        positions = positions + offsets
    positions = positions.flatten()
    num_tokens = positions.shape[0]
    cos_sin = cos_sin_cache.index_select(0, positions)
    cos, sin = cos_sin.chunk(2, dim=-1)

    query_shape = query.shape
    query = query.view(num_tokens, -1, head_size)
    query_rot = query[..., :rotary_dim]
    query_pass = query[..., rotary_dim:]
    query_rot = _apply_rotary_emb(query_rot, cos, sin, is_neox_style=is_neox_style)
    query = torch.cat((query_rot, query_pass), dim=-1).reshape(query_shape)

    key_shape = key.shape
    key = key.view(num_tokens, -1, head_size)
    key_rot = key[..., :rotary_dim]
    key_pass = key[..., rotary_dim:]
    key_rot = _apply_rotary_emb(key_rot, cos, sin, is_neox_style=is_neox_style)
    key = torch.cat((key_rot, key_pass), dim=-1).reshape(key_shape)
    return query, key


def _rotary_embedding_vllm_ref(
    positions: torch.Tensor,
    query: torch.Tensor,
    key: torch.Tensor,
    cos_sin_cache: torch.Tensor,
    rotary_dim: int,
    head_size: int,
    *,
    is_neox_style: bool = True,
    offsets: torch.Tensor | None = None,
) -> tuple[torch.Tensor, torch.Tensor]:
    """vLLM reference rotary_embedding impl."""
    from vllm import _custom_ops as vllm_custom_ops

    cos_sin_cache = cos_sin_cache.to(query.device, dtype=query.dtype)

    # vllm_custom_ops.rotary_embedding()/batched_rotary_embedding()
    # are in-place operations that update the query and key tensors.
    if offsets is not None:
        vllm_custom_ops.batched_rotary_embedding(
            positions, query, key, head_size, cos_sin_cache, is_neox_style, rotary_dim, offsets
        )
    else:
        vllm_custom_ops.rotary_embedding(positions, query, key, head_size, cos_sin_cache, is_neox_style)

    return query, key


def rotary_embedding(
    positions: torch.Tensor,
    query: torch.Tensor,
    key: torch.Tensor,
    cos_sin_cache: torch.Tensor,
    rotary_dim: int,
    head_size: int,
    *,
    is_neox_style: bool = True,
    offsets: torch.Tensor | None = None,
) -> tuple[torch.Tensor, torch.Tensor]:
    """Rotary embedding operation."""
    if envs.CONCH_ENABLE_VLLM and current_platform.has_cuda():
        return _rotary_embedding_vllm_ref(
            positions,
            query,
            key,
            cos_sin_cache,
            rotary_dim,
            head_size,
            is_neox_style=is_neox_style,
            offsets=offsets,
        )

    return _rotary_embedding_pytorch_ref(
        positions,
        query,
        key,
        cos_sin_cache,
        rotary_dim,
        head_size,
        is_neox_style=is_neox_style,
        offsets=offsets,
    )



# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

"""Environment variables."""

import os
from collections.abc import Callable
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    CONCH_BENCH_ENABLE_ALL_REF: bool
    CONCH_ENABLE_BNB: bool
    CONCH_ENABLE_VLLM: bool


environment_variables: dict[str, Callable[[], Any]] = {
    # Enable the vLLM/bitsandbytes reference implementations for all applicable benchmarks
    "CONCH_BENCH_ENABLE_ALL_REF": lambda: (
        os.environ.get("CONCH_BENCH_ENABLE_ALL_REF", "0").strip().lower() in ("1", "true")
    ),
    # Enable bitsandbytes kernels for testing/benchmarking
    "CONCH_ENABLE_BNB": lambda: (os.environ.get("CONCH_ENABLE_BNB", "0").strip().lower() in ("1", "true")),
    # Enable vLLM kernels for testing/benchmarking
    "CONCH_ENABLE_VLLM": lambda: (os.environ.get("CONCH_ENABLE_VLLM", "0").strip().lower() in ("1", "true")),
}


def __getattr__(name: str) -> Any:
    # lazy evaluation of environment variables
    if name in environment_variables:
        return environment_variables[name]()
    error_msg = f"module {__name__!r} has no attribute {name!r}"
    raise AttributeError(error_msg)


def __dir__() -> list[str]:
    return list(environment_variables.keys())



# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

"""Conch gemma_rms_norm benchmark."""

import sys
from typing import Final

import click
import torch

from conch.ops.normalization.gemma_rms_norm import gemma_rms_norm as gemma_rms_norm_conch
from conch.platforms import current_platform
from conch.reference.normalization.gemma_rms_norm import gemma_rms_norm as gemma_rms_norm_reference
from conch.third_party.vllm.utils import seed_everything
from conch.utils.benchmark import BenchmarkMetadata, benchmark_it


@click.command()
@click.option(
    "--embedding-size",
    required=False,
    type=int,
    default=2048,
    help="Embedding size",
)
@click.option(
    "--num-tokens",
    required=False,
    type=int,
    default=8192,
    help="Number of tokens",
)
@click.option(
    "--iteration-time-ms",
    required=False,
    type=int,
    default=10000,
    help="Time in milliseconds to run benchmark",
)
@click.option(
    "--warmup-time-ms",
    required=False,
    type=int,
    default=1000,
    help="Time in milliseconds to warmup before recording times",
)
@click.option(
    "--absolute-tolerance",
    required=False,
    type=float,
    default=1e-2,
    help="Absolute tolerance to match with",
)
@click.option(
    "--verbose",
    is_flag=True,
    help="Flag for printing verbose output",
)
@click.option(
    "--gpu",
    required=False,
    type=str,
    default=current_platform.device,
    help="Device to run on",
)
@click.option(
    "--csv",
    is_flag=True,
    help="Flag for printing results in CSV format",
)
def main(
    embedding_size: int,
    num_tokens: int,
    iteration_time_ms: int,
    warmup_time_ms: int,
    absolute_tolerance: float,
    verbose: bool,
    gpu: str,
    csv: bool,
) -> None:
    """Benchmark Conch GemmaRMSNorm op.

    Args:
        embedding_size: Embedding size.
        num_tokens: Number of tokens.
        iteration_time_ms: Time in milliseconds to run benchmark.
        warmup_time_ms: Time in milliseconds to warmup before recording times.
        absolute_tolerance: Absolute tolerance used to check accuracy of PyTorch vs. Conch.
        verbose: Flag to indicate whether or not to print verbose output.
        gpu: Which gpu to run on.
        csv: Flag for printing results in CSV format.
    """
    seed: Final = 0
    seed_everything(seed)

    device: Final = torch.device(gpu)
    torch.set_default_device(device)

    epsilon: Final = 1e-6

    metadata = BenchmarkMetadata(
        platform=current_platform.name(),
        params={
            "embedding_size": embedding_size,
            "num_tokens": num_tokens,
        },
    )

    x = torch.randn((num_tokens, embedding_size), dtype=torch.float16, device=device)
    weights = torch.randn((embedding_size,), device=device)

    x_ref = x.clone()
    x_conch = x.clone()

    result_ref = gemma_rms_norm_reference(x_ref, weights, epsilon, residual=None)
    result_conch = gemma_rms_norm_conch(x_conch, weights, epsilon, residual=None)

    # For mypy (if residual==None then result is single Tensor, not tuple[Tensor, Tensor])
    assert isinstance(result_ref, torch.Tensor)
    assert isinstance(result_conch, torch.Tensor)

    if not torch.allclose(result_ref, result_conch, atol=absolute_tolerance):
        print(f"WARNING: Reference and Conch results differ! (atol={absolute_tolerance})", file=sys.stderr)
        print(f"Output max diff: {(result_conch - result_ref).abs().max().item()}", file=sys.stderr)

        if verbose:
            print(f"Reference output: {result_ref}", file=sys.stderr)
            print(f"Conch output: {result_conch}", file=sys.stderr)
    else:
        print(f"Results matched with atol={absolute_tolerance} :)", file=sys.stderr)

    baseline_result = benchmark_it(
        lambda: gemma_rms_norm_reference(x_ref, weights, epsilon, residual=None),
        tag="Baseline",
        metadata=metadata,
        iteration_time_ms=iteration_time_ms,
        warmup_time_ms=warmup_time_ms,
    )

    conch_result = benchmark_it(
        lambda: gemma_rms_norm_conch(x_conch, weights, epsilon, residual=None),
        tag="Conch",
        metadata=metadata,
        iteration_time_ms=iteration_time_ms,
        warmup_time_ms=warmup_time_ms,
    )

    # Print results
    conch_result.print_parameters(csv=csv)
    conch_result.print_results(csv=csv)
    baseline_result.print_results(csv=csv)


if __name__ == "__main__":
    main()



# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

"""Bitsandbytes quantize PyTorch reference implementation."""

from collections.abc import Callable

import torch
import triton


def _pytorch_fp4_quantize(x: torch.Tensor, code: torch.Tensor | None) -> int:  # noqa: PLR0911
    """Quantize a floating point value to FP4."""
    assert x.ndim == 0, "Argument must be scalar tensor!"
    assert code is None

    sign = 8 if x < 0 else 0
    x = torch.abs(x)

    if x > 0.29166667:
        if x > 0.5833334:
            if x > 0.83333334:
                return 3 + sign
            return 2 + sign
        if x > 0.4166667:
            return 5 + sign
        return 4 + sign

    if x > 0.0859375:
        if x > 0.208333334:
            return 7 + sign
        return 6 + sign
    if x > 0.00260417:
        return 1 + sign
    return 0 + sign


def _pytorch_nf4_quantize(x: torch.Tensor, code: torch.Tensor | None) -> int:  # noqa: C901, PLR0911, PLR0912
    """Quantize a floating point value to NF4."""
    assert x.ndim == 0, "Argument must be scalar tensor!"
    assert code is None
    if x > 0.03979014977812767:
        if x > 0.3893125355243683:
            if x > 0.6427869200706482:
                if x > 0.8614784181118011:
                    return 15
                return 14
            if x > 0.5016634166240692:
                return 13
            return 12
        if x > 0.2035212516784668:
            if x > 0.2920137718319893:
                return 11
            return 10
        if x > 0.1202552504837513:
            return 9
        return 8
    if x > -0.33967943489551544:
        if x > -0.13791173323988914:
            if x > -0.045525018125772476:
                return 7
            return 6
        if x > -0.23460740596055984:
            return 5
        return 4
    if x > -0.6106329262256622:
        if x > -0.4599952697753906:
            return 3
        return 2
    if x > -0.8480964004993439:
        return 1
    return 0


def _pytorch_fp8_quantize(x: torch.Tensor, code: torch.Tensor | None) -> int:
    """Quantize a floating point value to FP8."""
    assert x.ndim == 0, "Argument must be scalar tensor!"
    assert code is not None
    pivot = 127
    upper_pivot = 255
    lower_pivot = 0

    lower = torch.tensor(-1.0)
    upper = torch.tensor(1.0)

    val = code[pivot]

    for i in [64, 32, 16, 8, 4, 2, 1]:
        if x > val:
            lower_pivot = pivot
            lower = val
            pivot += i
        else:
            upper_pivot = pivot
            upper = val
            pivot -= i

        val = code[pivot]

    if upper_pivot == 255:
        upper = code[upper_pivot]
    if lower_pivot == 0:
        lower = code[lower_pivot]

    if x > val:
        midpoint = (upper + val) * 0.5
        if x > midpoint:
            return upper_pivot
        return pivot

    midpoint = (lower + val) * 0.5
    if x < midpoint:
        return lower_pivot
    return pivot


def _get_pytorch_quant_method(quant_type: str) -> Callable[[torch.Tensor, torch.Tensor | None], int]:
    """Get quantization method."""
    if quant_type == "nf4":
        return _pytorch_nf4_quantize
    if quant_type == "fp4":
        return _pytorch_fp4_quantize
    return _pytorch_fp8_quantize


def quantize_blockwise_launcher(  # noqa: PLR0913
    x: torch.Tensor,
    absmax: torch.Tensor,
    out: torch.Tensor,
    code: torch.Tensor | None,
    blocksize: int,
    input_size: int,
    quant_type: str,
) -> tuple[torch.Tensor, torch.Tensor]:
    """Convert FP32/FP16/BF16 input tensor to NF4/FP4/FP8."""
    # This function assumes we are always packing two, NF4/FP4 elements into one 8-bit integer
    assert out.dtype == torch.uint8

    quant_method = _get_pytorch_quant_method(quant_type)

    num_blocks = triton.cdiv(input_size, blocksize)
    output_blocksize = blocksize // 2 if quant_type in ("fp4", "nf4") else blocksize

    # Process in blocks of size {blocksize}
    for block_index in range(num_blocks):
        # Get elements of current block
        input_block_offset = block_index * blocksize
        # Don't read extra elements if input_size is not perfectly divisible by blocksize
        this_block_size = min(blocksize, input_size - input_block_offset)
        block = x[input_block_offset : input_block_offset + this_block_size]

        # Reduce abs(max())
        local_absmax = torch.max(torch.abs(block))

        # Record for de-quantization
        absmax[block_index] = local_absmax

        # Invert so that we can _multiply_ by scaling factor during quantization
        local_absmax = 1.0 / local_absmax.to(torch.float32)

        output_block_offset = block_index * output_blocksize

        # Don't read extra elements if input_size is not perfectly divisible by blocksize
        this_output_block_size = this_block_size // 2 if quant_type in ("fp4", "nf4") else this_block_size

        # Pack two quantized elements into one for the output (two 4-bit values packed into a `uint8`)
        for element_index in range(this_output_block_size):
            if quant_type in ("fp4", "nf4"):
                # Initialize result to 0 (i.e. bitwise-or always flips the bit)
                result = 0
                # Pack first element (note: for parity with BNB, upcast to fp32)
                result |= quant_method(block[element_index * 2].to(torch.float32) * local_absmax, code) << 4
                # Pack second element (same upcasting as above)
                result |= quant_method(block[element_index * 2 + 1].to(torch.float32) * local_absmax, code)
            else:
                result = quant_method(block[element_index].to(torch.float32) * local_absmax, code)

            # Store result into output
            out[output_block_offset + element_index] = result

    return out, absmax



# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

"""Test cases for Triton int8 quant kernels."""

from typing import Final

import pytest
import torch

from conch.ops.quantization.int8 import scaled_int8_quant as scaled_int8_quant_triton
from conch.platforms import current_platform
from conch.reference.quantization.int8 import scaled_int8_quant as scaled_int8_quant_reference
from conch.third_party.vllm.utils import seed_everything

_DTYPES: Final = [torch.half, torch.bfloat16, torch.float]
_HIDDEN_SIZES: Final = [16, 67, 768, 5137, 8193]
_NUM_TOKENS: Final = [1, 7, 83, 4096]
_SEEDS: Final = [0]
_SCALE: Final = [0.1, 2.1]


@pytest.mark.parametrize("num_tokens", _NUM_TOKENS)
@pytest.mark.parametrize("hidden_size", _HIDDEN_SIZES)
@pytest.mark.parametrize("dtype", _DTYPES)
@pytest.mark.parametrize("seed", _SEEDS)
@pytest.mark.parametrize("scale", _SCALE)
@torch.inference_mode()
def test_static_scaled_int8_quant(
    num_tokens: int,
    hidden_size: int,
    dtype: torch.dtype,
    seed: int,
    scale: float,
) -> None:
    """Test Triton static_scaled_int8_quant vs. reference implementation."""
    seed_everything(seed)

    device: Final = torch.device(current_platform.device)
    torch.set_default_device(device)

    x = torch.rand(num_tokens, hidden_size, dtype=dtype, device=device) * 1000
    scale_arg = torch.tensor([scale], dtype=torch.float32, device=device)

    triton_output, _ = scaled_int8_quant_triton(x, scale_arg)

    ref_output = scaled_int8_quant_reference(x, scale_arg)

    # big atol to account for rounding errors
    torch.testing.assert_close(triton_output, ref_output, atol=1, rtol=0.0)



# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

"""Test Triton bnb dequantization implementation."""

from typing import Final

import pytest
import torch
import triton
import triton.language as tl

from conch.kernels.quantization.bitsandbytes.dequantize_blockwise import _fp4_dequantize, _nf4_dequantize
from conch.kernels.quantization.bitsandbytes.dequantize_blockwise import (
    dequantize_blockwise_launcher as dequantize_blockwise_triton,
)
from conch.ops.quantization.bitsandbytes.functional import (
    SUPPORTED_BLOCKSIZES,
    SUPPORTED_QUANT_TYPES,
    _create_dynamic_map,
)
from conch.platforms import current_platform
from conch.reference.quantization.bitsandbytes.dequantize_blockwise import (
    _pytorch_fp4_dequantize,
    _pytorch_nf4_dequantize,
)
from conch.reference.quantization.bitsandbytes.dequantize_blockwise import (
    dequantize_blockwise_launcher as dequantize_blockwise_reference,
)
from conch.third_party.vllm.utils import seed_everything

# Too many parameterizations makes the PyTorch-reference test cases too slow
_BLOCKSIZES_ABRIDGED: Final = [64, 1024]
_SIZE_MULTIPLIERS: Final = [2.5, 6]
_DTYPES: Final = [torch.float32, torch.float16, torch.bfloat16]


@triton.jit  # type: ignore[misc]
def _dequantize_fp4_kernel(x_ptr: tl.tensor, out_ptr: tl.tensor) -> None:
    val = tl.load(x_ptr)
    q_val = _fp4_dequantize(val)
    tl.store(out_ptr, q_val)


@triton.jit  # type: ignore[misc]
def _dequantize_nf4_kernel(x_ptr: tl.tensor, out_ptr: tl.tensor) -> None:
    val = tl.load(x_ptr)
    q_val = _nf4_dequantize(val)
    tl.store(out_ptr, q_val)


def _dequantize_launcher(val: int, quant_type: str) -> float:
    device = torch.device(current_platform.device)

    x = torch.full((1,), val, dtype=torch.uint8, device=device)
    out = torch.empty_like(x, dtype=torch.float32, device=device)

    dequantize_kernel = _dequantize_fp4_kernel if quant_type == "fp4" else _dequantize_nf4_kernel
    dequantize_kernel[(1,)](x, out)

    return out.cpu()[0].item()


@pytest.mark.parametrize(
    ("expected", "x"),
    [
        (-0.25, 15),
        (-0.166666, 14),
        (-0.5, 13),
        (-0.333333, 12),
        (-1.0, 11),
        (-0.6666666, 10),
        (-0.0052083333, 9),
        (-0.0, 8),
        (0.25, 7),
        (0.166666, 6),
        (0.5, 5),
        (0.333333, 4),
        (1.0, 3),
        (0.6666666, 2),
        (0.0052083333, 1),
        (0.0, 0),
    ],
)
def test_dequantize_fp4(expected: float, x: int) -> None:
    """Test FP4 mapping."""
    quant_type: Final = "fp4"
    assert _dequantize_launcher(x, quant_type) == pytest.approx(expected)
    assert _pytorch_fp4_dequantize(torch.tensor(x), None) == pytest.approx(expected)


@pytest.mark.parametrize(
    ("expected", "x"),
    [
        (1.00, 15),
        (0.7229568362236023, 14),
        (0.5626170039176941, 13),
        (0.44070982933044434, 12),
        (0.33791524171829224, 11),
        (0.24611230194568634, 10),
        (0.16093020141124725, 9),
        (0.07958029955625534, 8),
        (0.0, 7),
        (-0.09105003625154495, 6),
        (-0.18477343022823334, 5),
        (-0.28444138169288635, 4),
        (-0.39491748809814453, 3),
        (-0.5250730514526367, 2),
        (-0.6961928009986877, 1),
        (-1.0, 0),
    ],
)
def test_dequantize_nf4(expected: float, x: int) -> None:
    """Test NF4 mapping."""
    quant_type: Final = "nf4"
    assert _dequantize_launcher(x, quant_type) == pytest.approx(expected)
    assert _pytorch_nf4_dequantize(torch.tensor(x), None) == pytest.approx(expected)


@pytest.mark.parametrize("blocksize", _BLOCKSIZES_ABRIDGED)
@pytest.mark.parametrize("size_multiplier", _SIZE_MULTIPLIERS)
@pytest.mark.parametrize("dtype", _DTYPES)
@pytest.mark.parametrize("quant_type", SUPPORTED_QUANT_TYPES)
def test_dequantize(blocksize: int, size_multiplier: float, dtype: torch.dtype, quant_type: str) -> None:
    """Test blockwise dequantization method."""
    assert blocksize in SUPPORTED_BLOCKSIZES

    seed: Final = 0
    seed_everything(seed)

    device: Final = torch.device(current_platform.device)
    torch.set_default_device(device)

    output_size = int(blocksize * size_multiplier)
    num_blocks = triton.cdiv(output_size, blocksize)
    input_size = output_size if quant_type == "fp8" else output_size // 2

    x = torch.randint(0, 255, (input_size,), device=device, dtype=torch.uint8)
    absmax = torch.randn((num_blocks,), device=device, dtype=dtype)
    triton_out = torch.empty((output_size,), device=device, dtype=dtype)
    reference_out = torch.empty((output_size,), device=device, dtype=dtype)

    code = _create_dynamic_map() if quant_type == "fp8" else None

    triton_result = dequantize_blockwise_triton(
        x,
        absmax=absmax,
        out=triton_out,
        code=code,
        blocksize=blocksize,
        output_size=output_size,
        quant_type=quant_type,
    )

    reference_result = dequantize_blockwise_reference(
        x,
        absmax=absmax,
        out=reference_out,
        code=code,
        blocksize=blocksize,
        output_size=output_size,
        quant_type=quant_type,
    )

    torch.testing.assert_close(triton_result, reference_result)



# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

"""Test Triton bnb quantization implementation."""

from typing import Final

import pytest
import torch
import triton
import triton.language as tl

from conch.kernels.quantization.bitsandbytes.quantize_blockwise import _fp4_quantize, _nf4_quantize
from conch.kernels.quantization.bitsandbytes.quantize_blockwise import (
    quantize_blockwise_launcher as quantize_blockwise_triton,
)
from conch.ops.quantization.bitsandbytes.functional import (
    SUPPORTED_BLOCKSIZES,
    SUPPORTED_QUANT_TYPES,
    _create_dynamic_map,
    get_absmax_shape,
    get_quantized_output_shape,
)
from conch.platforms import current_platform
from conch.reference.quantization.bitsandbytes.quantize_blockwise import _pytorch_fp4_quantize, _pytorch_nf4_quantize
from conch.reference.quantization.bitsandbytes.quantize_blockwise import (
    quantize_blockwise_launcher as quantize_blockwise_reference,
)
from conch.third_party.vllm.utils import seed_everything

# Too many parameterizations makes the PyTorch-reference test cases too slow
_BLOCKSIZES_ABRIDGED: Final = [64, 1024]
_SIZE_MULTIPLIERS: Final = [2.5, 6]
_DTYPES: Final = [torch.float32, torch.float16, torch.bfloat16]


@triton.jit  # type: ignore[misc]
def _quantize_fp4_kernel(x_ptr: tl.tensor, out_ptr: tl.tensor) -> None:
    val = tl.load(x_ptr)
    q_val = _fp4_quantize(val)
    tl.store(out_ptr, q_val)


@triton.jit  # type: ignore[misc]
def _quantize_nf4_kernel(x_ptr: tl.tensor, out_ptr: tl.tensor) -> None:
    val = tl.load(x_ptr)
    q_val = _nf4_quantize(val)
    tl.store(out_ptr, q_val)


def _quantize_launcher(val: float, quant_type: str) -> int:
    device = torch.device(current_platform.device)

    x = torch.full((1,), val, dtype=torch.float32, device=device)
    out = torch.empty_like(x, dtype=torch.uint8, device=device)

    quantize_kernel = _quantize_fp4_kernel if quant_type == "fp4" else _quantize_nf4_kernel
    quantize_kernel[(1,)](x, out)

    return int(out.cpu()[0].item())


@pytest.mark.parametrize(
    ("x", "expected"),
    [
        (1.00, 15),
        (0.87, 15),
        (0.85, 14),
        (0.65, 14),
        (0.63, 13),
        (0.51, 13),
        (0.49, 12),
        (0.39, 12),
        (0.38, 11),
        (0.30, 11),
        (0.29, 10),
        (0.21, 10),
        (0.20, 9),
        (0.13, 9),
        (0.12, 8),
        (0.04, 8),
        (0.03, 7),
        (-0.04, 7),
        (-0.05, 6),
        (-0.13, 6),
        (-0.14, 5),
        (-0.23, 5),
        (-0.25, 4),
        (-0.33, 4),
        (-0.34, 3),
        (-0.45, 3),
        (-0.46, 2),
        (-0.60, 2),
        (-0.62, 1),
        (-0.84, 1),
        (-0.85, 0),
        (-1.0, 0),
    ],
)
def test_quantize_nf4(x: float, expected: int) -> None:
    """Test NF4 mapping."""
    quant_type: Final = "nf4"
    assert _quantize_launcher(x, quant_type) == expected
    assert _pytorch_nf4_quantize(torch.tensor(x), None) == expected


@pytest.mark.parametrize(
    ("x", "expected"),
    [
        (1.0, 3),
        (0.84, 3),
        (0.83, 2),
        (0.59, 2),
        (0.58, 5),
        (0.42, 5),
        (0.41, 4),
        (0.30, 4),
        (0.29, 7),
        (0.21, 7),
        (0.20, 6),
        (0.09, 6),
        (0.08, 1),
        (0.003, 1),
        (0.002, 0),
        (0.0, 0),
        (-0.000001, 8),
        (-0.002, 8),
        (-0.003, 9),
        (-0.08, 9),
        (-0.09, 14),
        (-0.20, 14),
        (-0.21, 15),
        (-0.29, 15),
        (-0.30, 12),
        (-0.41, 12),
        (-0.42, 13),
        (-0.58, 13),
        (-0.59, 10),
        (-0.83, 10),
        (-0.84, 11),
        (-1.0, 11),
    ],
)
def test_quantize_fp4(x: float, expected: int) -> None:
    """Test FP4 mapping."""
    quant_type: Final = "fp4"
    assert _quantize_launcher(x, quant_type) == expected
    assert _pytorch_fp4_quantize(torch.tensor(x), None) == expected


@pytest.mark.parametrize("blocksize", _BLOCKSIZES_ABRIDGED)
@pytest.mark.parametrize("size_multiplier", _SIZE_MULTIPLIERS)
@pytest.mark.parametrize("dtype", _DTYPES)
@pytest.mark.parametrize("quant_type", SUPPORTED_QUANT_TYPES)
def test_quantize_blockwise(blocksize: int, size_multiplier: float, dtype: torch.dtype, quant_type: str) -> None:
    """Test blockwise quantization method."""
    assert blocksize in SUPPORTED_BLOCKSIZES

    # TODO(jmanning):
    # Seeing odd AMD compiler error for this kernel for FP8 cases
    # E       RuntimeError: PassManager::run failed
    # .direnv/python-3.10.12/lib/python3.10/site-packages/triton/backends/amd/compiler.py:243: RuntimeError
    # loc("/home/$USER/conch/conch/kernels/quantization/bitsandbytes/quantize_blockwise.py":114:21): error: operand #1 does not dominate this use
    if current_platform.is_amd() and quant_type == "fp8":
        pytest.skip()

    # There are some small rounding discrepancies between Triton and PyTorch, and its tough to adjust tolerances
    # to account for it because we compre bitpacked results (so if there's a rounding discrepancy in what become
    # the higher-bits of the result, the tolerance would need to be much higher to account for it). Consider changing
    # this seed if test cases are failing
    seed: Final = 2
    seed_everything(seed)

    device: Final = torch.device(current_platform.device)
    torch.set_default_device(device)

    input_size = int(blocksize * size_multiplier)
    output_shape: Final = get_quantized_output_shape(input_size, quant_type)
    absmax_shape: Final = get_absmax_shape(input_size, blocksize)

    x = torch.randn((input_size,), device=device, dtype=dtype)
    if quant_type == "fp4":
        x = x.uniform_(-1.0, 1.0)

    reference_absmax = torch.empty(absmax_shape, device=device, dtype=dtype)
    reference_out = torch.empty(output_shape, device=device, dtype=torch.uint8)

    code = _create_dynamic_map() if quant_type == "fp8" else None

    quantize_blockwise_reference(
        x,
        reference_absmax,
        reference_out,
        code,
        blocksize,
        input_size,
        quant_type,
    )

    triton_absmax = torch.empty(absmax_shape, device=device, dtype=dtype)
    triton_out = torch.empty(output_shape, device=device, dtype=torch.uint8)

    quantize_blockwise_triton(
        x,
        triton_absmax,
        triton_out,
        code,
        blocksize,
        input_size,
        quant_type,
    )

    torch.testing.assert_close(reference_absmax, triton_absmax)
    torch.testing.assert_close(reference_out, triton_out, atol=1, rtol=5e-3)



# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

"""PyTorch reference implementation of vLLM copy_blocks."""

import torch

from conch import envs
from conch.platforms import current_platform


def _copy_blocks_pytorch_ref(
    key_caches: list[torch.Tensor], value_caches: list[torch.Tensor], block_mapping: list[tuple[int, int]]
) -> None:
    """Reference PyTorch-only implementation of copy_blocks."""
    for src, dst in block_mapping:
        for key_cache in key_caches:
            key_cache[dst].copy_(key_cache[src])
        for value_cache in value_caches:
            value_cache[dst].copy_(value_cache[src])


def _copy_blocks_vllm_ref(
    key_caches: list[torch.Tensor], value_caches: list[torch.Tensor], block_mapping: list[tuple[int, int]]
) -> None:
    """Reference vLLM implementation of copy_blocks."""
    from vllm._custom_ops import copy_blocks as copy_blocks_vllm

    block_mapping_tensor = torch.tensor(block_mapping, dtype=torch.int64, device=key_caches[0].device).view(-1, 2)
    copy_blocks_vllm(key_caches, value_caches, block_mapping_tensor)


def copy_blocks(
    key_caches: list[torch.Tensor],
    value_caches: list[torch.Tensor],
    block_mapping: list[tuple[int, int]],
) -> None:
    """Reference implementation of vLLM's copy_blocks operation."""
    if envs.CONCH_ENABLE_VLLM and current_platform.has_cuda():
        return _copy_blocks_vllm_ref(key_caches, value_caches, block_mapping)

    return _copy_blocks_pytorch_ref(key_caches, value_caches, block_mapping)



# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

"""Conch gelu_tanh_and_mul benchmark."""

import sys
from typing import Final

import click
import torch

from conch.ops.activation.gelu_tanh_and_mul import gelu_tanh_and_mul as gelu_tanh_and_mul_conch
from conch.platforms import current_platform
from conch.reference.activation.gelu_tanh_and_mul import gelu_tanh_and_mul as gelu_tanh_and_mul_reference
from conch.third_party.vllm.utils import seed_everything
from conch.utils.benchmark import BenchmarkMetadata, benchmark_it


@click.command()
@click.option(
    "--hidden-size",
    required=False,
    type=int,
    default=13824,
    help="Feedforward hidden size",
)
@click.option(
    "--num-tokens",
    required=False,
    type=int,
    default=8192,
    help="Number of tokens",
)
@click.option(
    "--iteration-time-ms",
    required=False,
    type=int,
    default=10000,
    help="Time in milliseconds to run benchmark",
)
@click.option(
    "--warmup-time-ms",
    required=False,
    type=int,
    default=1000,
    help="Time in milliseconds to warmup before recording times",
)
@click.option(
    "--absolute-tolerance",
    required=False,
    type=float,
    default=1e-3,
    help="Absolute tolerance to match with",
)
@click.option(
    "--verbose",
    is_flag=True,
    help="Flag for printing verbose output",
)
@click.option(
    "--gpu",
    required=False,
    type=str,
    default=current_platform.device,
    help="Device to run on",
)
@click.option(
    "--csv",
    is_flag=True,
    help="Flag for printing results in CSV format",
)
def main(
    hidden_size: int,
    num_tokens: int,
    iteration_time_ms: int,
    warmup_time_ms: int,
    absolute_tolerance: float,
    verbose: bool,
    gpu: str,
    csv: bool,
) -> None:
    """Benchmark Conch GeluTanhAndMul op.

    Args:
        hidden_size: Feedforward hidden size.
        num_tokens: Number of tokens.
        iteration_time_ms: Time in milliseconds to run the benchmark.
        warmup_time_ms: Time in milliseconds to warm up before recording times.
        absolute_tolerance: Absolute tolerance used to check accuracy of PyTorch vs. Conch.
        verbose: Flag to indicate whether or not to print verbose output.
        gpu: Which gpu to run on.
        csv: Flag for printing results in CSV format.
    """
    seed: Final = 0
    seed_everything(seed)

    device: Final = torch.device(gpu)
    torch.set_default_device(device)

    metadata = BenchmarkMetadata(
        platform=current_platform.name(),
        params={
            "hidden_size": hidden_size,
            "num_tokens": num_tokens,
        },
    )

    projections = torch.rand((num_tokens, hidden_size * 2), device=device)

    ref_output = gelu_tanh_and_mul_reference(projections)
    conch_output = gelu_tanh_and_mul_conch(projections)

    if not torch.allclose(ref_output, conch_output, atol=absolute_tolerance):
        print(f"WARNING: Reference and Conch results differ! (atol={absolute_tolerance})", file=sys.stderr)
        print(f"Output max diff: {(conch_output - ref_output).abs().max().item()}", file=sys.stderr)

        if verbose:
            print(f"Reference output: {ref_output}", file=sys.stderr)
            print(f"Conch output: {conch_output}", file=sys.stderr)
    else:
        print(f"Results matched with atol={absolute_tolerance} :)", file=sys.stderr)

    baseline_result = benchmark_it(
        lambda: gelu_tanh_and_mul_reference(projections),
        tag="Baseline",
        metadata=metadata,
        iteration_time_ms=iteration_time_ms,
        warmup_time_ms=warmup_time_ms,
    )

    conch_result = benchmark_it(
        lambda: gelu_tanh_and_mul_conch(projections),
        tag="Conch",
        metadata=metadata,
        iteration_time_ms=iteration_time_ms,
        warmup_time_ms=warmup_time_ms,
    )

    # Print results
    conch_result.print_parameters(csv=csv)
    conch_result.print_results(csv=csv)
    baseline_result.print_results(csv=csv)


if __name__ == "__main__":
    main()
